/*
 *
 *    OPEN-XCHANGE legal information
 *
 *    All intellectual property rights in the Software are protected by
 *    international copyright laws.
 *
 *
 *    In some countries OX, OX Open-Xchange, open xchange and OXtender
 *    as well as the corresponding Logos OX Open-Xchange and OX are registered
 *    trademarks.
 *    The use of the Logos is not covered by the GNU General Public License.
 *    Instead, you are allowed to use these Logos according to the terms and
 *    conditions of the Creative Commons License, Version 2.5, Attribution,
 *    Non-commercial, ShareAlike, and the interpretation of the term
 *    Non-commercial applicable to the aforementioned license is published
 *    on the web site http://www.open-xchange.com/EN/legal/index.html.
 *
 *    Please make sure that third-party modules and libraries are used
 *    according to their respective licenses.
 *
 *    Any modifications to this package must retain all copyright notices
 *    of the original copyright holder(s) for the original code used.
 *
 *    After any such modifications, the original and derivative code shall remain
 *    under the copyright of the copyright holder(s) and/or original author(s)per
 *    the Attribution and Assignment Agreement that can be located at
 *    http://www.open-xchange.com/EN/developer/. The contributing author shall be
 *    given Attribution for the derivative code and a license granting use.
 *
 *     Copyright (C) 2016 OX Software GmbH
 *     Mail: info@open-xchange.com
 *
 *
 *     This program is free software; you can redistribute it and/or modify it
 *     under the terms of the GNU General Public License, Version 2 as published
 *     by the Free Software Foundation.
 *
 *     This program is distributed in the hope that it will be useful, but
 *     WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *     or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 *     for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc., 59
 *     Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

/**
 * @author sven.jacobi@open-xchange.com
 */

package com.openexchange.office.odf.styles;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Random;
import java.util.Stack;
import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;
import org.xml.sax.SAXException;
import com.openexchange.office.odf.AttributesImpl;
import com.openexchange.office.odf.DLNode;
import com.openexchange.office.odf.INodeAccessor;
import com.openexchange.office.odf.Namespaces;
import com.openexchange.office.odf.OpAttrs;
import com.openexchange.office.odf.properties.FontFace;
import com.openexchange.office.odf.properties.GraphicProperties;
import com.openexchange.office.odf.properties.MapProperties;
import com.openexchange.office.odf.properties.ParagraphProperties;
import com.openexchange.office.odf.properties.TextProperties;

public class StyleManager {

	private Styles 				styles;
	private final FamilyStyles 	familyStyles = new FamilyStyles();
	private MasterStyles		masterStyles;
	private final FamilyStyles 	familyMasterStyles = new FamilyStyles();
	private AutomaticStyles 	automaticStylesMaster;
	private final FamilyStyles 	familyAutomaticStylesMaster = new FamilyStyles();
	private AutomaticStyles 	automaticStylesContent;
	private final FamilyStyles 	familyAutomaticStylesContent = new FamilyStyles();
	private FontFaceDecls		fontFaceDeclsMaster;
	private FontFaceDecls		fontFaceDeclsContent;

	private static Random		random;
	private boolean				hasTabsRelativeToIndent = false;

	/*
	 * The FamilyStyles class is a wrapper for a two key hash map, the
	 * first key is the family and the second key is the name of the style
	 */
	private class FamilyStyles {

		private final HashMap<String, HashMap<String, DLNode<Object>>> stylesByFamily = new HashMap<String, HashMap<String, DLNode<Object>>>();

		// adds the style, family and name is taken from the style
		public void addStyle(DLNode<Object> styleBaseNode) {
			final StyleBase styleBase = (StyleBase)styleBaseNode.getObject();
			HashMap<String, DLNode<Object>> familyMap = stylesByFamily.get(styleBase.getFamily());
			if(familyMap==null) {
				familyMap = new HashMap<String, DLNode<Object>>();
				stylesByFamily.put(styleBase.getFamily(), familyMap);
			}
			familyMap.put(styleBase.getName(), styleBaseNode);
		}

		public DLNode<Object> removeStyle(String family, String name) {
			return stylesByFamily.get(family).remove(name);
		}

		public StyleBase getStyle(String family, String name) {
			if(name==null||name.isEmpty()) {
				return null;
			}
			final HashMap<String, DLNode<Object>> styles = stylesByFamily.get(family);
			if(styles==null) {
				return null;
			}
			final DLNode<Object> styleBaseNode = styles.get(name);
			return styleBaseNode!=null ? (StyleBase)styleBaseNode.getObject() : null;
		}

		public HashMap<String, DLNode<Object>> getStylesByFamily(String family) {
			return stylesByFamily.get(family);
		}
	}

	private final HashMap<StyleBase, String> allStyles = new HashMap<StyleBase, String>();

	public void addStyle(StyleBase styleBase) {
		final DLNode<Object> styleBaseNode = new DLNode<Object>(styleBase);
		if(styleBase.isMasterStyle()) {
			masterStyles.getContent().addNode(styleBaseNode);
			familyMasterStyles.addStyle(styleBaseNode);
		}
		else if(styleBase.isAutoStyle()) {
			if(styleBase.isContentStyle()) {
				automaticStylesContent.getContent().addNode(styleBaseNode);
				familyAutomaticStylesContent.addStyle(styleBaseNode);
			}
			else {
				automaticStylesMaster.getContent().addNode(styleBaseNode);
				familyAutomaticStylesMaster.addStyle(styleBaseNode);
			}
		}
		else {
			styles.getContent().addNode(styleBaseNode);
			familyStyles.addStyle(styleBaseNode);
		}
		allStyles.put(styleBase, styleBase.getName());
	}

	public StyleBase removeStyle(String family, String name, boolean autoStyle, boolean masterStyle) {
		DLNode<Object> styleBaseNode = null;
		if(autoStyle) {
			if(masterStyle) {
				styleBaseNode = familyAutomaticStylesMaster.removeStyle(family, name);
				if(styleBaseNode!=null) {
					automaticStylesMaster.getContent().removeNode(styleBaseNode);
				}
			}
			else {
				styleBaseNode = familyAutomaticStylesContent.removeStyle(family, name);
				if(styleBaseNode!=null) {
					automaticStylesContent.getContent().removeNode(styleBaseNode);
				}
			}
		}
		else if(masterStyle) {
			styleBaseNode = familyMasterStyles.removeStyle(family, name);
			if(styleBaseNode!=null) {
				masterStyles.getContent().removeNode(styleBaseNode);
			}
		}
		else {
			styleBaseNode = familyStyles.removeStyle(family, name);
			if(styleBaseNode!=null) {
				styles.getContent().removeNode(styleBaseNode);
			}
		}
		if(styleBaseNode==null) {
			return null;
		}
		final StyleBase styleBase = (StyleBase)styleBaseNode.getObject();
		final String allStyleName = allStyles.get(styleBase);
		if(allStyleName.equals(styleBase.getName())) {
			allStyles.remove(styleBase);
		}
		return styleBase;
	}

	public StyleBase getAutoStyle(String name, String family, boolean content) {
		final FamilyStyles familyStyles = content ? familyAutomaticStylesContent : familyAutomaticStylesMaster;
		return familyStyles.getStyle(family, name);
	}

	public String getUniqueStyleName(String family, boolean content) {
		int i = getRandom();
		while(true) {
			final String name = family.charAt(0) + Integer.toString(i);
			if(getStyle(name, family, content)==null) {
				return name;
			}
			i++;
		}
	}

	private int getRandom() {
		if(random==null) {
			random = new Random();
		}
		return Math.abs(random.nextInt());
	}

	public StyleBase getStyle(String name, String family, boolean content) {
		StyleBase style = getAutoStyle(name, family, content);
		if(style==null) {
			style = familyStyles.getStyle(family, name);
		}
		return style;
	}

	public Styles getStyles() {
		return styles;
	}

	public MasterStyles getMasterStyles() {
		return masterStyles;
	}

	public AutomaticStyles getAutomaticStyles(boolean content) {
		if(content) {
			return automaticStylesContent;
		}
		else {
			return automaticStylesMaster;
		}
	}

	public FontFaceDecls getFontFaceDecls(boolean content) {
		if(content) {
			return fontFaceDeclsContent;
		}
		else {
			return fontFaceDeclsMaster;
		}
	}

    public StyleBase getStyleByDisplayName(String displayName, String family, boolean content) {
        FamilyStyles autofamilyStyles = content ? familyAutomaticStylesContent : familyAutomaticStylesMaster;

        for (DLNode<Object> node : autofamilyStyles.getStylesByFamily(family).values()) {
            StyleBase style = (StyleBase) node.getObject();
            if (displayName.equals(style.getDisplayName())) {
                return style;
            }
        }

        for (DLNode<Object> node : familyStyles.getStylesByFamily(family).values()) {
            StyleBase style = (StyleBase) node.getObject();
            if (displayName.equals(style.getDisplayName())) {
                return style;
            }
        }
        return null;
    }

	// ------------------------------------------------------------------------------------------------------------------
	// document type can be "text", "spreadsheet" or "presentation"
	public void createInsertStyleOperations(INodeAccessor nodeAccessor, String documentType, JSONArray operationsQueue)
		throws JSONException {

		final HashSet<String> createdStyles = new HashSet<String>();
		final Iterator<Object> styleIter = nodeAccessor.getContent().iterator();
		while(styleIter.hasNext()) {
			final Object o = styleIter.next();
			if(o instanceof StyleBase) {
				final StyleBase style = (StyleBase)o;
				switch(style.getFamily()) {
					case "paragraph" : {
						if(documentType.equals("text")) {
							_createInsertStyleOperation(style, "paragraph", createdStyles, operationsQueue);
						}
						break;
					}
					case "text" : {
						if(documentType.equals("text")) {
							_createInsertStyleOperation(style, "character", createdStyles, operationsQueue);
						}
						break;
					}
					case "graphic" : {
						if(documentType.equals("text")) {
							_createInsertStyleOperation(style, "drawing", createdStyles, operationsQueue);
						}
						break;
					}
					case "table": {
						if(documentType.equals("text")) {
							_createInsertStyleOperation(style, "table", createdStyles, operationsQueue);
						}
						break;
					}
					case "table-cell" : {
						if(documentType.equals("spreadsheet")) {
							_createInsertStyleOperation(style, "cell", createdStyles, operationsQueue);
						}
						break;
					}
					case "list-style" : {
						if(documentType.equals("text")) {
							_createInsertStyleOperation(style, "list-style", createdStyles, operationsQueue);
						}
						break;
					}
				}
			}
		}
	}

	public void createInsertAutoListStylesOperations(boolean content, JSONArray operationsQueue)
		throws JSONException {
		final FamilyStyles familyStyles = content ? familyAutomaticStylesContent : familyAutomaticStylesMaster;
		final HashMap<String, DLNode<Object>> stylesByFamily = familyStyles.getStylesByFamily("list-style");
		if(stylesByFamily!=null) {
			final Iterator<DLNode<Object>> styleIter = stylesByFamily.values().iterator();
			final HashSet<String> createdStyles = new HashSet<String>();
			while(styleIter.hasNext()) {
				final StyleBase style = (StyleBase)styleIter.next().getObject();
				if(style.getFamily().equals("list-style")) {
					_createInsertStyleOperation(style, "list-style", createdStyles, operationsQueue);
				}
			}
		}
	}

	private void _createInsertStyleOperation(StyleBase style, String destType, HashSet<String> createdStyles, JSONArray operationsQueue)
		throws JSONException {

		// not creating styles twice... parents are inserted first
		final String styleIdentifier = style.getFamily() + style.getName() + style.isAutoStyle();
		if(createdStyles.contains(styleIdentifier)) {
			return;
		}
		createdStyles.add(styleIdentifier);

		final StyleBase parentStyle = familyStyles.getStyle(style.getFamily(), style.getParent());
		if(parentStyle!=null) {
			_createInsertStyleOperation(parentStyle, destType, createdStyles, operationsQueue);
		}
		// ongoing with style ...
		final JSONObject op = new JSONObject();
		if(destType.equals("list-style")) {
			op.put("name", "insertListStyle");
			op.put("listStyleId", style.getName());
		}
		else {
			if(style.isAutoStyle()) {
				op.put("name", "insertAutoStyle");
			}
			else {
				op.put("name", "insertStyleSheet");
			}
			op.put("type", destType);
			op.put("styleId", style.getName());
		}
		final String displayName = style.getDisplayName();
		if(displayName!=null&&!displayName.isEmpty()) {
			op.put("styleName", displayName);
		}
		if (style.isDefaultStyle()) {
            op.put("default", style.isDefaultStyle());
        }
		final OpAttrs attrs = new OpAttrs();
		style.createAttrs(this, attrs);
        if(parentStyle!=null) {
        	if(style.isAutoStyle()) {
        		attrs.put("styleId", parentStyle.getName());
        	}
        	else {
        		op.put("parent", parentStyle.getName());
        	}
        }
        else if(!style.isDefaultStyle()) {
        	op.put("parent", "_default");
        }
		if(!attrs.isEmpty()) {
			if(destType.equals("list-style")) {
				op.put("listDefinition", attrs);
			}
			else {
				op.put("attrs", attrs);
			}
		}

/*
        if (outlineLevel != null || nextStyleId != null) {
            JSONObject paraProps;
            if (componentProps.containsKey("paragraph")) {
                paraProps = (JSONObject) componentProps.get("paragraph");
            } else {
                paraProps = new JSONObject();
                componentProps.put("paragraph", paraProps);
            }
            if (outlineLevel != null) {
                paraProps.put("outlineLevel", outlineLevel - 1);
            }
            if (nextStyleId != null && !nextStyleId.isEmpty()) {
                paraProps.put("nextStyleId", nextStyleId);
            }
            componentProps.put("paragraph", paraProps);
            insertComponentObject.put("attrs", componentProps);
        }
*/
        operationsQueue.put(op);
	}

	private static String styleTypeToFamily(String styleType) {
		switch(styleType) {
			case "page" 	: return "page-layout";	// text style
			case "drawing" 	: return "graphic"; 	// text style
			case "character": return "text";		// text style
			case "paragraph": return "paragraph";	// text style
			case "table"	: return "table";		// text style
			case "cell"		: return "table-cell"; 	// spreadsheet style
			default: throw new RuntimeException("odf: unknown style type used");
		}
	}

	public void insertStyleSheet(String styleType, String styleId, String styleName, JSONObject attrs, String parent, boolean isDefaultStyle, boolean hidden)
		throws JSONException {

		final StyleBase styleBase = createStyleBase(styleTypeToFamily(styleType), styleId, false, false);
		styleBase.setIsDefaultStyle(isDefaultStyle);
		if(styleName!=null&&!styleName.isEmpty()) {
			styleBase.setDisplayName(styleName);
		}
		if(parent!=null&&!parent.isEmpty()&&!parent.equals("_default")) {
			styleBase.setParent(parent);
		}
		styleBase.applyAttrs(this, attrs);
		addStyle(styleBase);
	}

    public void insertAutoStyle(String styleType, String styleId, JSONObject attrs, boolean isDefaultStyle)
    	throws JSONException {

		final StyleBase styleBase = createStyleBase(styleTypeToFamily(styleType), styleId, true, true);
		styleBase.setIsDefaultStyle(isDefaultStyle);
		styleBase.applyAttrs(this, attrs);

        final String parent = attrs != null ? attrs.optString("styleId") : null;
        if (parent != null && !parent.isEmpty() && !parent.equals("_default")) {
            styleBase.setParent(parent);
        }

		addStyle(styleBase);
    }

    public void deleteAutoStyle(String styleType, String name) {
    	final String family = styleTypeToFamily(styleType);
    	removeStyle(family, name, true, false);
    }

    public void changeAutoStyle(String type, String styleId, JSONObject attrs)
    	throws JSONException, SAXException {

    	deleteAutoStyle(type, styleId);
    	insertAutoStyle(type, styleId, attrs, false);
    }

    public String applyAttributes(String family, String styleId, boolean contentAutoStyle, JSONObject attrs)
		throws JSONException {

		final StyleBase currentStyleBase;
		if(styleId!=null&&!styleId.isEmpty()) {
			final StyleBase originalStyleBase = getStyle(styleId, family, contentAutoStyle);
			if(originalStyleBase!=null) {
				currentStyleBase = originalStyleBase.clone();
				currentStyleBase.setName(getUniqueStyleName(family, contentAutoStyle));
			}
			else {
				currentStyleBase = createStyleBase(family, styleId, true, contentAutoStyle);
			}
		}
		else {
			currentStyleBase = createStyleBase(family, getUniqueStyleName(family, contentAutoStyle), true, contentAutoStyle);
		}

		// if a new style is applied, then we need to merge each parent auto style first
		final Object oStyleId = attrs.opt("styleId");
		if(oStyleId!=null) {
			final List<StyleBase> parentAutoStyles = new ArrayList<StyleBase>();
			StyleBase p = currentStyleBase;
			do {
				final StyleBase parentStyle = findParentAutoStyle(p, parentAutoStyles);
				if(parentStyle!=null) {
					parentAutoStyles.add(parentStyle);
				}
				p = parentStyle;
			}
			while(p!=null);

			for(int i=0; i<parentAutoStyles.size(); i++) {
				currentStyleBase.mergeAttrs(parentAutoStyles.get(i));
			}
			if(oStyleId instanceof String) {
				currentStyleBase.setParent((String)oStyleId);
			}
			else if (oStyleId==JSONObject.NULL) {
				currentStyleBase.setParent(null);
			}
		}
		currentStyleBase.setIsAutoStyle(true);
		currentStyleBase.setIsContentStyle(contentAutoStyle);

		currentStyleBase.applyAttrs(this, attrs);

		final String existingStyleId = getExistingStyleIdForStyleBase(currentStyleBase);
		if(existingStyleId!=null) {
			return existingStyleId;
		}
		else {
			/* TODO: isEmpty implementieren
			if(currentStyleBase.isEmpty()) {
				return currentStyleBase.getParent();
			}
			*/
			this.addStyle(currentStyleBase);
			return currentStyleBase.getName();
		}
	}

	// we will never change existing autostyles, so it is safe to cache used autoStyle attributes
	private final HashMap<String, HashMap<String, Object>> cachedAutoStyleMasterAttributes = new HashMap<String, HashMap<String, Object>>();
	private final HashMap<String, HashMap<String, Object>> cachedAutoStyleContentAttributes = new HashMap<String, HashMap<String, Object>>();;

	public void createAutoStyleAttributes(OpAttrs attrs, String styleId, String family, boolean contentStyle) {

		final HashMap<String, HashMap<String, Object>> cachedAutoStyles
			= contentStyle ? cachedAutoStyleContentAttributes : cachedAutoStyleMasterAttributes;

		if(styleId!=null&&!styleId.isEmpty()) {
			StyleBase style = getAutoStyle(styleId, family, contentStyle);
			if(style!=null) {
				final Stack<StyleBase> autoStyleStack = new Stack<StyleBase>();
				HashMap<String, Object> cachedStyle = null;
				while(style!=null) {
					// only adding autoStyles to the stack which are not already cached
					cachedStyle = cachedAutoStyles.get(style.getName());
					if(cachedStyle!=null) {
						break;
					}
					autoStyleStack.push(style);
					style = findParentAutoStyle(style, autoStyleStack);
				}
				if(cachedStyle!=null) {
					deepCopy(cachedStyle, attrs);	// we have cached attributes (containing also the correct style)
				}
				else {
					// the last one in the stack has the correct style
					styleId = autoStyleStack.peek().getParent();
					if(styleId!=null&&!styleId.isEmpty()) {
						attrs.put("styleId", styleId);
					}
				}
				while(!autoStyleStack.isEmpty()) {	// copy attributes from autostyles that not have been cached already
					style = autoStyleStack.pop();
					style.createAttrs(this, attrs);
					cachedStyle = new HashMap<String, Object>();
					deepCopy(attrs, cachedStyle);
					cachedAutoStyles.put(style.getName(), cachedStyle);
				}
			}
			else {	// this must be the correct style because there exists no auto style with this id
				attrs.put("styleId", styleId);
			}
		}
	}

	private final HashMap<String, HashMap<String, Object>> cachedParagraphPropertiesMaster = new HashMap<String, HashMap<String, Object>>();
	private final HashMap<String, HashMap<String, Object>> cachedParagraphPropertiesContent = new HashMap<String, HashMap<String, Object>>();;

	public void collectAllParagraphPropertiesFromStyle(OpAttrs attrs, String styleId, String family, boolean contentStyle) {

		final HashMap<String, HashMap<String, Object>> cachedProperties
			= contentStyle ? cachedParagraphPropertiesContent : cachedParagraphPropertiesMaster;

		if(styleId!=null&&!styleId.isEmpty()) {
			StyleBase style = getAutoStyle(styleId, family, contentStyle);
			if(style!=null) {
				final Stack<StyleBase> styleStack = new Stack<StyleBase>();
				HashMap<String, Object> cacheHit = null;
				while(style!=null) {
					// only adding autoStyles to the stack which are not already cached
					cacheHit = cachedProperties.get(style.getName());
					if(cacheHit!=null) {
						break;
					}
					styleStack.push(style);
					style = findParentStyle(style, styleStack);
				}
				if(cacheHit!=null) {
					deepCopy(cacheHit, attrs);	// we have cached attributes
				}
				while(!styleStack.isEmpty()) {	// copy attributes from autostyles that not have been cached already
					style = styleStack.pop();
					if(style instanceof IParagraphProperties) {
						final ParagraphProperties paragraphProperties = ((IParagraphProperties)style).getParagraphProperties();
						paragraphProperties.createAttrs(this, contentStyle, attrs);
						cacheHit = new HashMap<String, Object>();
						deepCopy(attrs, cacheHit);
						cachedProperties.put(style.getName(), cacheHit);
					}
				}
			}
		}
	}

	private final HashMap<String, HashMap<String, Object>> cachedTextPropertiesMaster = new HashMap<String, HashMap<String, Object>>();
	private final HashMap<String, HashMap<String, Object>> cachedTextPropertiesContent = new HashMap<String, HashMap<String, Object>>();;

	public void collectAllTextPropertiesFromStyle(OpAttrs attrs, String styleId, String family, boolean contentStyle) {

		final HashMap<String, HashMap<String, Object>> cachedProperties
			= contentStyle ? cachedTextPropertiesContent : cachedTextPropertiesMaster;

		if(styleId!=null&&!styleId.isEmpty()) {
			StyleBase style = getAutoStyle(styleId, family, contentStyle);
			if(style!=null) {
				final Stack<StyleBase> styleStack = new Stack<StyleBase>();
				HashMap<String, Object> cacheHit = null;
				while(style!=null) {
					// only adding autoStyles to the stack which are not already cached
					cacheHit = cachedProperties.get(style.getName());
					if(cacheHit!=null) {
						break;
					}
					styleStack.push(style);
					style = findParentStyle(style, styleStack);
				}
				if(cacheHit!=null) {
					deepCopy(cacheHit, attrs);	// we have cached attributes
				}
				while(!styleStack.isEmpty()) {	// copy attributes from autostyles that not have been cached already
					style = styleStack.pop();
					if(style instanceof ITextProperties) {
						final TextProperties TextProperties = ((ITextProperties)style).getTextProperties();
						TextProperties.createAttrs(this, contentStyle, attrs);
						cacheHit = new HashMap<String, Object>();
						deepCopy(attrs, cacheHit);
						cachedProperties.put(style.getName(), cacheHit);
					}
				}
			}
		}
	}

	private final HashMap<String, HashMap<String, Object>> cachedDrawingPropertiesMaster = new HashMap<String, HashMap<String, Object>>();
	private final HashMap<String, HashMap<String, Object>> cachedDrawingPropertiesContent = new HashMap<String, HashMap<String, Object>>();;

	public void collectAllGraphicPropertiesFromStyle(OpAttrs attrs, String styleId, String family, boolean contentStyle) {

		final HashMap<String, HashMap<String, Object>> cachedProperties
			= contentStyle ? cachedDrawingPropertiesContent : cachedDrawingPropertiesMaster;

		if(styleId!=null&&!styleId.isEmpty()) {
			StyleBase style = getAutoStyle(styleId, family, contentStyle);
			if(style!=null) {
				final Stack<StyleBase> styleStack = new Stack<StyleBase>();
				HashMap<String, Object> cacheHit = null;
				while(style!=null) {
					// only adding autoStyles to the stack which are not already cached
					cacheHit = cachedProperties.get(style.getName());
					if(cacheHit!=null) {
						break;
					}
					styleStack.push(style);
					style = findParentStyle(style, styleStack);
				}
				if(cacheHit!=null) {
					deepCopy(cacheHit, attrs);	// we have cached attributes
				}
				while(!styleStack.isEmpty()) {	// copy attributes from autostyles that not have been cached already
					style = styleStack.pop();
					if(style instanceof IGraphicProperties) {
						final GraphicProperties graphicProperties = ((IGraphicProperties)style).getGraphicProperties();
						graphicProperties.createAttrs(this, contentStyle, attrs);
						cacheHit = new HashMap<String, Object>();
						deepCopy(attrs, cacheHit);
						cachedProperties.put(style.getName(), cacheHit);
					}
				}
			}
		}
	}

    public String applyDataStyle(String code, int id, boolean volatileStyle, boolean automaticStyle, boolean contentStyle) {

    	final String dataStyleFamily;
        switch(id) {
            case 1:  code = "0"; dataStyleFamily = "number-style"; break;
            case 2:  code = "0.00"; dataStyleFamily = "number-style"; break;
            case 3:  code = "#,##0"; dataStyleFamily = "number-style"; break;
            case 4:  code = "#,##0.00"; dataStyleFamily = "number-style"; break;
            case 5:  code = "#,##0[$$-409]"; dataStyleFamily = "currency-style"; break;
            case 6:  code = "#,##0[$$-409];[RED]-#,##0[$$-409]"; dataStyleFamily = "currency-style"; break;
            case 7:  code = "#,##0.00[$$-409]"; dataStyleFamily = "currency-style"; break;
            case 8:  code = "#,##0.00[$$-409];[RED]-#,##0.00[$$-409]"; dataStyleFamily = "currency-style"; break;
            case 9:  code = "0%"; dataStyleFamily = "percentage-style"; break;
            case 10: code = "0.00%"; dataStyleFamily = "percentage-style"; break;
            case 11: code = "0.00E+00"; dataStyleFamily = "number-style"; break;
            case 12: code = "# ?/?"; dataStyleFamily = "number-style"; break;
            case 13: code = "# ??/??"; dataStyleFamily = "number-style"; break;
            case 14:  code = "MM/DD/YYYY"; dataStyleFamily = "date-style"; break;
            case 15:  code = "D-MMM-YY"; dataStyleFamily = "date-style"; break;
            case 16:  code = "D-MMM"; dataStyleFamily = "date-style"; break;
            case 17:  code = "MMM-YY"; dataStyleFamily = "date-style"; break;
            case 18:  code = "H:MM AM/PM"; dataStyleFamily = "date-style"; break;
            case 19:  code = "H:MM:SS AM/PM"; dataStyleFamily = "date-style"; break;
            case 20:  code = "H:MM"; dataStyleFamily = "time-style"; break;
            case 21:  code = "H:MM:SS"; dataStyleFamily = "time-style"; break;
            case 22:  code = "M/D/YYYY H:MM"; dataStyleFamily = "date-style"; break;
            case 23:  // PASSTHROUGH INTENTED !!
            case 24:  // PASSTHROUGH INTENTED !!
            case 25:  // PASSTHROUGH INTENTED !!
            case 26:  code = "0"; dataStyleFamily = "number-style"; break;
            case 27:  // PASSTHROUGH INTENTED !!
            case 28:  // PASSTHROUGH INTENTED !!
            case 29:  // PASSTHROUGH INTENTED !!
            case 30:  // PASSTHROUGH INTENTED !!
            case 31:  code = "M/D/YYYY"; dataStyleFamily = "date-style"; break;
            case 32:  // PASSTHROUGH INTENTED !!
            case 33:  // PASSTHROUGH INTENTED !!
            case 34:  // PASSTHROUGH INTENTED !!
            case 35:  code = "HH:MM:SS"; dataStyleFamily = "time-style"; break;
            case 36:  code = "M/D/YYYY"; dataStyleFamily = "date-style"; break;
            case 37:  code = "#,##0_);(#,##0)"; dataStyleFamily = "number-style"; break;
            case 38:  code = "#,##0_);[RED](#,##0)"; dataStyleFamily = "number-style"; break;
            case 39:  code = "#,##0.00_);(#,##0.00)"; dataStyleFamily = "number-style"; break;
            case 40:  code = "#,##0.00_);[RED](#,##0.00)"; dataStyleFamily = "number-style"; break;
            case 41:  code = "_-* #,##0 _\u20ac_-;-* #,##0 _\u20ac_-;_-* \"-\" _\u20ac_-;_-@_-"; dataStyleFamily = "number-style"; break;
            case 42:  code = "_-* #,##0 \"\u20ac\"_-;-* #,##0 \"\u20ac\"_-;_-* \"-\" \"\u20ac\"_-;_-@_-"; dataStyleFamily = "number-style"; break;
            case 43:  code = "_-* #,##0.00 _\u20ac_-;-* #,##0.00 _\u20ac_-;_-* \"-\"?? _\u20ac_-;_-@_-"; dataStyleFamily = "number-style"; break;
            case 44:  code = "_-* #,##0.00 \"\u20ac\"_-;-* #,##0.00 \"\u20ac\"_-;_-* \"-\"?? \"\u20ac\"_-;_-@_-"; dataStyleFamily = "number-style"; break;
            case 45:  code = "MM:SS"; dataStyleFamily = "time-style"; break;
            case 46:  code = "[h]:mm:ss"; dataStyleFamily = "time-style"; break;
            case 47:  code = "mm:ss.0"; dataStyleFamily = "time-style"; break;
            case 48:  code = "##0.0E+0"; dataStyleFamily = "number-style"; break;
            case 49:  code = "@"; dataStyleFamily = "text-style"; break;
            case 50:  code = "M/D/YYYY"; dataStyleFamily = "date-style"; break;
            case 51:  // PASSTHROUGH INTENTED !!
            case 52:  // PASSTHROUGH INTENTED !!
            case 53:  // PASSTHROUGH INTENTED !!
            case 54:  // PASSTHROUGH INTENTED !!
            case 55:  // PASSTHROUGH INTENTED !!
            case 56:  // PASSTHROUGH INTENTED !!
            case 57:  // PASSTHROUGH INTENTED !!
            case 58:  code = "M/D/YYYY"; dataStyleFamily = "date-style"; break;
            case 59:  code = "0"; dataStyleFamily = "number-style"; break;
            case 60:  code = "0.00"; dataStyleFamily = "number-style"; break;
            case 61:  code = "#,##0"; dataStyleFamily = "number-style"; break;
            case 62:  code = "#,##0.00"; dataStyleFamily = "number-style"; break;
            case 63:  code = "#,##0 \"\u20ac\";-#,##0 \"\u20ac\""; dataStyleFamily = "number-style"; break;
            case 64:  code = "#,##0 \"\u20ac\";[RED]-#,##0 \"\u20ac\""; dataStyleFamily = "number-style"; break;
            case 65:  code = "#,##0.00 \"\u20ac\";-#,##0.00 \"\u20ac\""; dataStyleFamily = "number-style"; break;
            case 66:  code = "#,##0.00 \"\u20ac\";[RED]-#,##0.00 \"\u20ac\""; dataStyleFamily = "number-style"; break;
            case 67:  code = "0%"; dataStyleFamily = "percentage-style"; break;
            case 68:  code = "0.00%"; dataStyleFamily = "percentage-style"; break;
            case 69:  code = "0.00E+00"; dataStyleFamily = "number-style"; break;
            case 70:  code = "# ?/?"; dataStyleFamily = "number-style"; break;
            case 71:  // PASSTHROUGH INTENTED !!
            case 72:  code = "M/D/YYYY"; dataStyleFamily = "date-style"; break;
            case 73:  code = "D-MMM-YY"; dataStyleFamily = "date-style"; break;
            case 74:  code = "D-MMM"; dataStyleFamily = "date-style"; break;
            case 75:  code = "MMM-YY"; dataStyleFamily = "date-style"; break;
            case 76:  code = "HH:MM"; dataStyleFamily = "time-style"; break;
            case 77:  code = "HH:MM:SS"; dataStyleFamily = "time-style"; break;
            case 78:  code = "M/D/YYYY H:MM"; dataStyleFamily = "time-style"; break;
            case 79:  code = "mm:ss"; dataStyleFamily = "time-style"; break;
            case 80:  code = "[h]:mm:ss"; dataStyleFamily = "time-style"; break;
            case 81:  code = "mm:ss.0"; dataStyleFamily = "time-style"; break;
            default: {
                String compareCode = code.replaceAll("AM", "xx9999xx");
                compareCode = compareCode.replaceAll("PM", "xx9999xx");
                //remove quoted parts
                compareCode = removeQuotedAndColor(compareCode);
                if(compareCode.contains("@")) {
                	dataStyleFamily = "text-style";
                } else if(compareCode.equals("BOOLEAN")) {
                	dataStyleFamily = "boolean-style";
                } else if(compareCode.contains("[$") ) {
                	dataStyleFamily = "currency-style";
                } else if(compareCode.contains("%") ) {
                	dataStyleFamily = "percentage-style";
                } else if(compareCode.contains("D")|| compareCode.contains("d") || compareCode.contains("M") || compareCode.contains("Y") || compareCode.contains("y") ||
                        compareCode.contains("WW") || compareCode.contains("ww") ) {
                	dataStyleFamily = "date-style";
                } else if(compareCode.contains("h") || compareCode.contains("H") ||compareCode.contains("m") || compareCode.contains("s") || compareCode.contains("S") ||
                         compareCode.contains("xx9999xx")) {
                	dataStyleFamily = "time-style";
                } else {
                	dataStyleFamily = "number-style";
                }
            }
        }

    	final NumberStyleBase numberStyleBase = (NumberStyleBase)createStyleBase(dataStyleFamily, getUniqueStyleName("data-style", true), automaticStyle, contentStyle);
    	if(volatileStyle) {
    		numberStyleBase.setIsVolatile(true);
    	}
        if(id==14) {
        	numberStyleBase.getAttributes().setValue(Namespaces.NUMBER, "automatic-order", "number:automatic-order", "true");
        	numberStyleBase.getAttributes().setValue(Namespaces.NUMBER, "format-source", "number:format-source", null);
        }
    	final ArrayList<String> partArray = new ArrayList<String>();
    	String numberFormatCode = code;
        while(numberFormatCode.contains(";")) {
            int partStart = numberFormatCode.lastIndexOf(";", numberFormatCode.length());
            String part = numberFormatCode.substring(partStart + 1);
            partArray.add(0, part);
            numberFormatCode = numberFormatCode.substring(0, partStart);
        }
        partArray.add(0, numberFormatCode);
        numberStyleBase.setFormat(partArray.get(partArray.size()-1));
        if(partArray.size()>1) {
            for(int partIndex=0;partIndex<partArray.size()-1;++partIndex) {
            	final String partStyleName = applyDataStyle(partArray.get(partIndex), -1, true, automaticStyle, contentStyle);
                //generate an appropriate condition and add this style as sub style and make part style a volatile style
            	String condition = "value()";
                // TODO: extract condition if available (e.g. '[>5]')
                condition += partArray.size() == 1  ? ">=0" : partIndex == 0 ? ">0"  : "<0";
                final MapProperties mapProperties = new MapProperties(new AttributesImpl());
                mapProperties.getAttributes().setValue(Namespaces.STYLE, "apply-style-name", "style:apply-style-name", partStyleName);
                mapProperties.getAttributes().setValue(Namespaces.STYLE, "condition", "style:condition", condition);
                numberStyleBase.getMapStyleList().add(mapProperties);
            }
        }
        final String existingStyleId = getExistingStyleIdForStyleBase(numberStyleBase);
		if(existingStyleId!=null) {
			return existingStyleId;
		}
		else {
			addStyle(numberStyleBase);
			return numberStyleBase.getName();
		}
    }

    public String getExistingStyleIdForStyleBase(StyleBase styleBase) {
    	return allStyles.get(styleBase);
    }

    private static String removeQuotedAndColor(String compareCode) {
        while(compareCode.contains("\"")) {
            int firstQuote = compareCode.indexOf("\"");
            if(firstQuote == compareCode.length() - 1) {
                break;
            }
            int secondQuote = compareCode.indexOf("\"", firstQuote + 1);
            if(secondQuote < 0){
                break;
            }
            String tmp = compareCode.substring(0, firstQuote);
            tmp += compareCode.substring(secondQuote + 1);
            compareCode = tmp;
        }
        int openBracket = compareCode.indexOf("[");
        boolean hasCurrency = false;
        while(openBracket >= 0){
            int closeBracket = compareCode.indexOf("]", openBracket);
            if(closeBracket > openBracket){
                String innerText = compareCode.substring(openBracket + 1, closeBracket);
                if(innerText.startsWith("$") && innerText.length() > 1) {
                    hasCurrency = true;
                }
                compareCode = compareCode.substring(0, openBracket) + compareCode.substring(closeBracket + 1);
            }
            openBracket = compareCode.indexOf("[");
        }
        if(hasCurrency) {
            compareCode += "[$]"; //reinsert currency marker
        }
        return compareCode;
    }

	private StyleBase findParentStyle(StyleBase style, Iterable<StyleBase> autoStyleStack) {
		final StyleBase parent = getStyle(style.getParent(), style.getFamily(), style.isContentStyle());
		if(parent!=null&&autoStyleStack!=null) {
			final Iterator<StyleBase> iter = autoStyleStack.iterator();
			while(iter.hasNext()) {
				if(parent.getName().equals(iter.next().getName())) {
					return null;
				}
			}
		}
		return parent;
	}

	/* returns the parent autostyle or zero. If autoStyleStack is given then the parent is checked
	   against the autoStyleStack to prevent recursion
	 */
	private StyleBase findParentAutoStyle(StyleBase style, Iterable<StyleBase> autoStyleStack) {
		final StyleBase parent = getAutoStyle(style.getParent(), style.getFamily(), style.isContentStyle());
		if(parent!=null&&autoStyleStack!=null) {
			final Iterator<StyleBase> iter = autoStyleStack.iterator();
			while(iter.hasNext()) {
				if(parent.getName().equals(iter.next().getName())) {
					return null;
				}
			}
		}
		return parent;
	}

	// is copying the content of each hashmap into dest.
	@SuppressWarnings({ "rawtypes", "unchecked" })
	public static void deepCopy(Map<String, Object> source, HashMap<String, Object> dest) {
		final Iterator<Entry<String, Object>> sourceIter = source.entrySet().iterator();
		while(sourceIter.hasNext()) {
			final Entry<String, Object> sourceEntry = sourceIter.next();
			if(sourceEntry.getValue() instanceof Map) {
				Object o = dest.get(sourceEntry.getKey());
				if(!(o instanceof HashMap)) {
					o = new HashMap<String, Object>();
					dest.put(sourceEntry.getKey(), o);
				}
				deepCopy((Map)sourceEntry.getValue(), (HashMap)o);
			}
			else {
				dest.put(sourceEntry.getKey(), sourceEntry.getValue());
			}
		}
	}

	// Nice to know, if a frame uses borders instead of stroke depends on having "Graphics"
	// or "Frame" as parent style The style itself can be empty, it is just the name that matters.
	public void updateFrameStyleProperty(StyleBase styleBase) {
		StyleBase parent = styleBase;
		Stack<StyleBase> styleStack = null;
		while(parent!=null&&!parent.getName().equals("Graphics")&&!parent.getName().equals("Frame")) {
			if(!(parent instanceof IGraphicProperties)) {
				return;
			}
			final Boolean isFrameOrGraphic = ((IGraphicProperties)parent).getGraphicProperties().getIsFrameOrGraphic();
			if(isFrameOrGraphic!=null) {
				((IGraphicProperties)styleBase).getGraphicProperties().setIsFrameOrGraphic(isFrameOrGraphic);
				if(styleStack!=null) {
					final Iterator<StyleBase> styleStackIter = styleStack.iterator();
					while(styleStackIter.hasNext()) {
						((IGraphicProperties)styleStackIter.next()).getGraphicProperties().setIsFrameOrGraphic(isFrameOrGraphic);
					}
				}
				return;
			}
			if(styleStack==null) {
				styleStack = new Stack<StyleBase>();
			}
			styleStack.push(parent);

			parent = findParentStyle(parent, styleStack);
		}
		if(parent!=null&&parent instanceof IGraphicProperties) {
			((IGraphicProperties)parent).getGraphicProperties().setIsFrameOrGraphic(true);
		}
		if(styleStack!=null) {
			final Iterator<StyleBase> styleStackIter = styleStack.iterator();
			while(styleStackIter.hasNext()) {
				((IGraphicProperties)styleStackIter.next()).getGraphicProperties().setIsFrameOrGraphic(parent!=null);
			}
		}
	}

	private StyleBase createStyleBase(String family, String styleId, boolean automaticStyle, boolean contentStyle) {
		switch(family) {
			case "text" : {
				return new StyleText(styleId, automaticStyle, contentStyle);
			}
			case "paragraph" : {
				return new StyleParagraph(styleId, automaticStyle, contentStyle);
			}
			case "section" : {
				return new StyleSection(styleId, automaticStyle, contentStyle);
			}
			case "ruby" : {
				return new StyleRuby(styleId, automaticStyle, contentStyle);
			}
			case "table" : {
				return new StyleTable(styleId, automaticStyle, contentStyle);
			}
			case "table-column" : {
				return new StyleTableColumn(styleId, automaticStyle, contentStyle);
			}
			case "table-row" : {
				return new StyleTableRow(styleId, automaticStyle, contentStyle);
			}
			case "table-cell" : {
				return new StyleTableCell(styleId, automaticStyle, contentStyle);
			}
			case "graphic" : {
				return new StyleGraphic(styleId, automaticStyle, contentStyle);
			}
			case "presentation" : {
				return new StylePresentation(styleId, automaticStyle, contentStyle);
			}
			case "drawing-page" : {
				return new StyleDrawingPage(styleId, automaticStyle, contentStyle);
			}
			case "chart" : {
				return new StyleChart(styleId, automaticStyle, contentStyle);
			}
			case "page-layout" : {
				return new StylePageLayout(styleId, automaticStyle, contentStyle);
			}
			case "number-style": {
	    		return new NumberNumberStyle(styleId, automaticStyle, contentStyle);
			}
			case "boolean-style": {
	    		return new NumberBooleanStyle(styleId, automaticStyle, contentStyle);
			}
			case "currency-style" : {
	    		return new NumberCurrencyStyle(styleId, automaticStyle, contentStyle);
			}
			case "date-style" : {
	    		return new NumberDateStyle(styleId, automaticStyle, contentStyle);
			}
			case "percentage-style" : {
	    		return new NumberPercentageStyle(styleId, automaticStyle, contentStyle);
			}
			case "text-style" : {
	    		return new NumberTextStyle(styleId, automaticStyle, contentStyle);
			}
			case "time-style" : {
	    		return new NumberTimeStyle(styleId, automaticStyle, contentStyle);
			}
			case "list-style" : {
	    		return new TextListStyle(styleId, automaticStyle, contentStyle);
			}
			case "stroke-dash" : {
				return new StrokeDashStyle(styleId, automaticStyle, contentStyle);
			}
		}
		return null;
	}

	public boolean hasTabsRelativeToIndent() {
		return hasTabsRelativeToIndent;
	}

	public void setTabsRelativeToIndent(boolean hasTabsRelativeToIndent) {
		this.hasTabsRelativeToIndent = hasTabsRelativeToIndent;
	}

    public void addFontToDocument(String fontName) {
    	if(fontName==null||fontName.isEmpty()) {
    		return;
    	}
    	if(fontFaceDeclsMaster==null||fontFaceDeclsContent==null) {
    		return;
    	}
    	if(fontFaceDeclsMaster.getFontFaces().containsKey(fontName)) {
    		return;
    	}
    	switch(fontName) {
    		case "Andale Mono": {
	            insertFontDescription("Andale Mono", null, "Andale Mono", null, null, null);
	            break;
    		}
    		case "Arial":  {
	            insertFontDescription("Arial", null, "Arial", "swiss", "variable", null);
	            break;
    		}
    		case "Book Antiqua": {
	            insertFontDescription("Book Antiqua", null, "Book Antiqua", "roman", "variable", null);
	            break;
    		}
    		case "Calibri": {
	            insertFontDescription("Calibri", null, "Calibri", "swiss", "variable", "2 15 5 2 2 2 4 3 2 4");
	            break;
    		}
    		case "Cambria": {
	            insertFontDescription("Cambria", null, "Cambria", "roman", "variable", "2 4 5 3 5 4 6 3 2 4");
	            break;
    		}
    		case "Consolas": {
	            insertFontDescription("Consolas", null, "Consolas", "modern", "fixed", null);
	            break;
    		}
    		case "Courier New": {
	            insertFontDescription("Courier New", null, "Courier New", "modern", "fixed", null);
	            break;
    		}
    		case "Courier": {
	            insertFontDescription("Courier", null, "Courier", "modern", "fixed", null);
	            break;
    		}
    		case "Georgia": {
	            insertFontDescription("Georgia", null, "Georgia", "roman", "variable", null);
	            break;
    		}
    		case "Helvetica": {
	            insertFontDescription("Helvetica", null, "Helvetica", "swiss", null, null);
	            break;
    		}
    		case "Impact": {
	            insertFontDescription("Impact", null, "Impact", "swiss", "variable", null);
	            break;
    		}
    		case "Mangal": {
	            insertFontDescription("Mangal", null, "Mangal", "system", "variable", null);
	            break;
    		}
    		case "Mangal1": {
	            insertFontDescription("Mangal1", null, "Mangal", null, null, null);
	            break;
    		}
    		case "Microsoft YaHei": {
	            insertFontDescription("Microsoft YaHei", null, "Microsoft YaHei", "system", "variable", null);
	            break;
    		}
    		case "MS Gothic": {
	            insertFontDescription("MS Gothic", null, "MS Gothic", "modern", "fixed", "2 11 6 9 7 2 5 8 2 4");
	            break;
    		}
    		case "MS Mincho": {
	            insertFontDescription("MS Mincho", null, "MS Mincho", "modern", "fixed", "2 2 6 9 4 2 5 8 3 4");
	            break;
    		}
    		case "Palatino": {
	            insertFontDescription("Palatino", null, "Palatino", "roman", null, null);
	            break;
    		}
    		case "SimSun": {
	            insertFontDescription("SimSun", null, "SimSun", "system", "variable", null);
	            break;
    		}
    		case "Tahoma": {
	            insertFontDescription("Tahoma", null, "Tahoma", "swiss", "variable", null);
	            break;
    		}
    		case "Times New Roman": {
	            insertFontDescription("Times New Roman", null, "Times New Roman", "roman", "variable", "2 2 6 3 5 4 5 2 3 4");
	            break;
    		}
    		case "Times": {
	            insertFontDescription("Times", null, "Times", "roman", null, null);
	            break;
    		}
    		case "Verdana": {
	            insertFontDescription("Verdana", null, "Verdana", "swiss", "variable", null);
	            break;
	        }
    	}
    }

    public void insertFontDescription(String fontName, String[] altNames, String family, String familyGeneric, String pitch, String panose1) {
    	if(fontName!=null&&!fontName.isEmpty()) {
	        final FontFace fontFace = new FontFace(new AttributesImpl());
	        final AttributesImpl fontFaceAttributes = fontFace.getAttributes();
	        fontFaceAttributes.setValue(Namespaces.STYLE, "name", "style:name", fontName);
	        if(family!=null&&!family.isEmpty()) {
	        	fontFaceAttributes.setValue(Namespaces.SVG, "font-family", "svg:font-family", family);
	        }
	        if(familyGeneric!= null&&!familyGeneric.isEmpty()) {
	        	fontFaceAttributes.setValue(Namespaces.STYLE, "font-family-generic", "style:font-family-generic", familyGeneric);
	        }
	        if(pitch!=null&&!pitch.isEmpty()) {
	            fontFaceAttributes.setValue(Namespaces.STYLE, "font-pitch", "style:font-pitch", pitch);
	        }
	        if(panose1!=null&&!panose1.isEmpty()) {
	            if(panose1.contains("[")) {
	                panose1 = panose1.substring(1, panose1.length() - 1);
	            }
	            if(panose1.contains(",")) {
	                panose1 = panose1.replace(',', ' ');
	            }
	            fontFaceAttributes.setValue(Namespaces.SVG, "panose1", "svg:panose1", panose1);
	        }
	        fontFaceDeclsMaster.getFontFaces().put(fontName, fontFace);
	        fontFaceDeclsContent.getFontFaces().put(fontName, fontFace.clone());
    	}
    }

// ------------------------------------------------------------------------------------------------------------------
// Following methods are to be called from the corresponding SaxContextHandlers only to initialize the StyleManager -
// Styles have to be inserted into the DOM otherwise style changes arent't saved when writing the document.         -
// ------------------------------------------------------------------------------------------------------------------

	public void setStyles(Styles styles) {
		this.styles = styles;
	}

	public void setMasterStyles(MasterStyles masterStyles) {
		this.masterStyles = masterStyles;
	}

	public void setAutomaticStyles(AutomaticStyles automaticStyles, boolean content) {
		if(content) {
			this.automaticStylesContent = automaticStyles;
		}
		else {
			this.automaticStylesMaster = automaticStyles;
		}
	}

	public void setFontFaceDecls(FontFaceDecls fontFaceDecls, boolean content) {
		if(content) {
			fontFaceDeclsContent = fontFaceDecls;
		}
		else {
			fontFaceDeclsMaster = fontFaceDecls;
		}
	}
}
