/*
 *
 *    OPEN-XCHANGE legal information
 *
 *    All intellectual property rights in the Software are protected by
 *    international copyright laws.
 *
 *
 *    In some countries OX, OX Open-Xchange, open xchange and OXtender
 *    as well as the corresponding Logos OX Open-Xchange and OX are registered
 *    trademarks.
 *    The use of the Logos is not covered by the GNU General Public License.
 *    Instead, you are allowed to use these Logos according to the terms and
 *    conditions of the Creative Commons License, Version 2.5, Attribution,
 *    Non-commercial, ShareAlike, and the interpretation of the term
 *    Non-commercial applicable to the aforementioned license is published
 *    on the web site http://www.open-xchange.com/EN/legal/index.html.
 *
 *    Please make sure that third-party modules and libraries are used
 *    according to their respective licenses.
 *
 *    Any modifications to this package must retain all copyright notices
 *    of the original copyright holder(s) for the original code used.
 *
 *    After any such modifications, the original and derivative code shall remain
 *    under the copyright of the copyright holder(s) and/or original author(s)per
 *    the Attribution and Assignment Agreement that can be located at
 *    http://www.open-xchange.com/EN/developer/. The contributing author shall be
 *    given Attribution for the derivative code and a license granting use.
 *
 *     Copyright (C) 2016 OX Software GmbH
 *     Mail: info@open-xchange.com
 *
 *
 *     This program is free software; you can redistribute it and/or modify it
 *     under the terms of the GNU General Public License, Version 2 as published
 *     by the Free Software Foundation.
 *
 *     This program is distributed in the hope that it will be useful, but
 *     WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *     or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 *     for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc., 59
 *     Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

package com.openexchange.office.ooxml.drawingml;

import java.util.HashMap;
import java.util.List;
import java.util.logging.Logger;
import javax.xml.bind.annotation.XmlElements;
import org.docx4j.dml.CTEffectList;
import org.docx4j.dml.CTLineJoinRound;
import org.docx4j.dml.CTLineProperties;
import org.docx4j.dml.CTNoFillProperties;
import org.docx4j.dml.CTRegularTextRun;
import org.docx4j.dml.CTShapeProperties;
import org.docx4j.dml.CTSolidColorFillProperties;
import org.docx4j.dml.CTTextBody;
import org.docx4j.dml.CTTextBodyProperties;
import org.docx4j.dml.CTTextCharacterProperties;
import org.docx4j.dml.CTTextField;
import org.docx4j.dml.CTTextLineBreak;
import org.docx4j.dml.CTTextParagraph;
import org.docx4j.dml.CTTextParagraphProperties;
import org.docx4j.dml.ITextCharacterProperties;
import org.docx4j.dml.STLineCap;
import org.docx4j.dml.chart.CTArea3DChart;
import org.docx4j.dml.chart.CTAreaChart;
import org.docx4j.dml.chart.CTAxDataSource;
import org.docx4j.dml.chart.CTAxPos;
import org.docx4j.dml.chart.CTBar3DChart;
import org.docx4j.dml.chart.CTBarChart;
import org.docx4j.dml.chart.CTBarDir;
import org.docx4j.dml.chart.CTBarSer;
import org.docx4j.dml.chart.CTBoolean;
import org.docx4j.dml.chart.CTBubbleChart;
import org.docx4j.dml.chart.CTBubbleSer;
import org.docx4j.dml.chart.CTCatAx;
import org.docx4j.dml.chart.CTChart;
import org.docx4j.dml.chart.CTChartLines;
import org.docx4j.dml.chart.CTChartSpace;
import org.docx4j.dml.chart.CTDLbls;
import org.docx4j.dml.chart.CTDPt;
import org.docx4j.dml.chart.CTDateAx;
import org.docx4j.dml.chart.CTDoughnutChart;
import org.docx4j.dml.chart.CTFirstSliceAng;
import org.docx4j.dml.chart.CTGapAmount;
import org.docx4j.dml.chart.CTHoleSize;
import org.docx4j.dml.chart.CTLayout;
import org.docx4j.dml.chart.CTLegend;
import org.docx4j.dml.chart.CTLegendPos;
import org.docx4j.dml.chart.CTLine3DChart;
import org.docx4j.dml.chart.CTLineChart;
import org.docx4j.dml.chart.CTLineSer;
import org.docx4j.dml.chart.CTMarker;
import org.docx4j.dml.chart.CTNumData;
import org.docx4j.dml.chart.CTNumDataSource;
import org.docx4j.dml.chart.CTNumRef;
import org.docx4j.dml.chart.CTNumVal;
import org.docx4j.dml.chart.CTOfPieChart;
import org.docx4j.dml.chart.CTOrientation;
import org.docx4j.dml.chart.CTOverlap;
import org.docx4j.dml.chart.CTPie3DChart;
import org.docx4j.dml.chart.CTPieChart;
import org.docx4j.dml.chart.CTPlotArea;
import org.docx4j.dml.chart.CTRadarChart;
import org.docx4j.dml.chart.CTRadarStyle;
import org.docx4j.dml.chart.CTScaling;
import org.docx4j.dml.chart.CTScatterChart;
import org.docx4j.dml.chart.CTScatterSer;
import org.docx4j.dml.chart.CTScatterStyle;
import org.docx4j.dml.chart.CTSerTx;
import org.docx4j.dml.chart.CTStockChart;
import org.docx4j.dml.chart.CTStrData;
import org.docx4j.dml.chart.CTStrRef;
import org.docx4j.dml.chart.CTStrVal;
import org.docx4j.dml.chart.CTStyle;
import org.docx4j.dml.chart.CTSurface3DChart;
import org.docx4j.dml.chart.CTSurfaceChart;
import org.docx4j.dml.chart.CTTickLblPos;
import org.docx4j.dml.chart.CTTickMark;
import org.docx4j.dml.chart.CTTitle;
import org.docx4j.dml.chart.CTTx;
import org.docx4j.dml.chart.CTUnsignedInt;
import org.docx4j.dml.chart.CTValAx;
import org.docx4j.dml.chart.IAxisDescription;
import org.docx4j.dml.chart.IListSer;
import org.docx4j.dml.chart.ISerContent;
import org.docx4j.dml.chart.ObjectFactory;
import org.docx4j.dml.chart.STAxPos;
import org.docx4j.dml.chart.STLegendPos;
import org.docx4j.dml.chart.STOrientation;
import org.docx4j.dml.chart.STRadarStyle;
import org.docx4j.dml.chart.STScatterStyle;
import org.docx4j.dml.chart.STTickLblPos;
import org.docx4j.dml.chart.STTickMark;
import org.docx4j.dml.chart.layer.NoAxIdChart;
import org.docx4j.dml.chartStyle2012.CTChartStyle;
import org.docx4j.dml.chartStyle2012.CTColorStyle;
import org.docx4j.dml.chartStyle2012.CTColorStyleVariation;
import org.docx4j.dml.chartStyle2012.CTStyleEntry;
import org.docx4j.dml.chartStyle2012.CTStyleReference;
import org.docx4j.jaxb.Context;
import org.docx4j.openpackaging.exceptions.InvalidFormatException;
import org.docx4j.openpackaging.exceptions.PartUnrecognisedException;
import org.docx4j.openpackaging.parts.ThemePart;
import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;
import com.openexchange.office.ooxml.tools.Commons;

final public class DMLChartSpace {

	private static final JSONArray EMPTYTEXTARRAY;
	private static final CTShapeProperties DEFAULTSHAPE;

	static{
	    EMPTYTEXTARRAY = new JSONArray();
	    EMPTYTEXTARRAY.put(" ");

	    DEFAULTSHAPE = new CTShapeProperties();
	    DEFAULTSHAPE.setSolidFill(new CTSolidColorFillProperties());
	}

    public static void createChartSpaceOperations(JSONArray operationsArray, JSONArray position, ChartWrapper chartPart) throws Exception {

		final CTChartSpace chartSpace = chartPart.getChartSpace();

		// MSO2012 chartStyles can be zero...
		final CTChartStyle chartStyle = chartPart.getChartStyle();
		// final CTColorStyle chartColorStyle = chartPart.getChartColorStyle();
		final ThemePart themePart = chartPart.getThemePart();

		final CTChart chart = chartSpace.getChart();
		if (chart == null) {
			return;
		}
		final CTPlotArea plotArea = chart.getPlotArea();
		if (plotArea != null) {
            final List<IListSer> chartList = plotArea.getAreaChartOrArea3DChartOrLineChart();
			// TODO: what if there is more than one chart in this list ?
			if (!chartList.isEmpty()) {
                final IListSer chartSer = chartList.get(0);

				CTBoolean varyColors = chartSer.getVaryColors();

				final JSONObject insertDrawingOperation = operationsArray.getJSONObject(operationsArray.length() - 1);
				boolean valueLabel = isDataLabel(chartSer);
				boolean curved = isCurved(chartSer);

				// we are taking only the first chart ...
				for (int seriesIndex = 0; seriesIndex < chartSer.getSer().size(); seriesIndex++) {
					final ISerContent ser = chartSer.getSer().get(seriesIndex);

					// title
					final JSONObject series = new JSONObject();
					final CTSerTx tx = ser.getTx();
					Object title = null;
					if (tx != null) {
						if (tx.getStrRef() != null) {
							title = tx.getStrRef().getF();
                        }
                        else {
							title = new JSONArray();
							((JSONArray) title).put(tx.getV());
						}
					}
					if (title != null) {
						series.put("title", title);
					}

					// names
					Object names = null;
					final CTAxDataSource cat = ser.getCat();
					if (cat != null) {
						if (cat.getStrRef() != null) {
							names = cat.getStrRef().getF();
						}
						else if (cat.getStrLit() != null) {
							final CTStrData strLit = cat.getStrLit();
							final List<CTStrVal> strLitPtList = strLit.getPt();
							if (strLitPtList != null && !strLitPtList.isEmpty()) {
								names = new JSONArray();
								for (final CTStrVal strLitPtEntry : strLitPtList) {
									if (strLitPtEntry != null) {
										((JSONArray) names).put(strLitPtEntry.getV());
                                    }
                                    else {
										((JSONArray) names).put((String) null);
									}
								}
							}
						}
						else if (cat.getNumRef() != null) {
							names = cat.getNumRef().getF();
						}
						else if (cat.getNumLit() != null) {
							final CTNumData numLit = cat.getNumLit();
							final List<CTNumVal> numLitPtList = numLit.getPt();
							if (numLitPtList != null && !numLitPtList.isEmpty()) {
								names = new JSONArray();
								for (final CTNumVal numLitPtEntry : numLitPtList) {
									if (numLitPtEntry != null) {
										((JSONArray) names).put(numLitPtEntry.getV());
									}
									else {
										((JSONArray) names).put((String) null);
									}
								}
							}
						}
						else if (cat.getMultiLvlStrRef() != null) {
							names = cat.getMultiLvlStrRef().getF();
						}
					}
					if (names != null) {
						series.put("names", names);
					}

					// values
					final CTNumDataSource numDataSource = ser.getVal();
					final Object values = getData(numDataSource);

					if (values != null) {
						series.put("values", values);
					}

					if (ser instanceof CTBubbleSer) {
						final CTBubbleSer serBB = (CTBubbleSer) ser;

						final CTNumDataSource bubblesizes = serBB.getBubbleSize();
						if (bubblesizes != null) {
							final Object bubbles = getData(bubblesizes);
							if (bubbles != null) {
								series.put("bubbles", bubbles);
							}
						}
					}
					final CTMarker marker = ser.getMarker();
                    if (marker != null && marker.getSymbol() != null && marker.getSymbol().getVal() != null) {
						series.put("marker", marker.getSymbol().getVal().value());
					}


					if (series.length() == 0) {
						//TODO: log the broken series?
						continue;
					}

					final JSONObject attrs = Commons.surroundJSONObject("series", series);
                    CTShapeProperties shape = ser.getSpPr();

                    if (shape == null) {
                        if (chartStyle.getDataPoint().getFillRef().getIdx() == 0) {

                            //TODO: without fallback the whole series would be invisible, we need DataPoint-Operations to fix tha problem

                            if (ser.getDPt().size()>0) {
                                varyColors = getBoolean(true);
                                shape = DEFAULTSHAPE;
                            } else {
                                Logger.getAnonymousLogger().warning("series has no datapoints, cannot create fallback for invisible data series!!!");
                            }
                        }
                    }

					DMLHelper.createJsonFromShapeProperties(attrs, shape, themePart, chartStyle.getDataPoint(), chartPart.getChartPart(), false, true);

                    if (ser instanceof CTLineSer || ser instanceof CTScatterSer) {
                        JSONObject line = attrs.optJSONObject("line");
                        if (line != null) {
                            if (!line.has("type")) {
                                line.put("type", "solid");
                            }
                        }
                        attrs.remove("fill");
                    }

                    if (marker != null) {
                        final CTShapeProperties shapePr = marker.getSpPr();
                        if (shapePr != null) {
                            final CTStyleEntry styleRef = chartStyle.getDataPointMarker();
                            final CTStyleReference fillRef = styleRef != null ? styleRef.getFillRef() : null;
                            DMLHelper.createJsonFromFillProperties(attrs, marker.getSpPr(), themePart, fillRef, chartPart.getChartPart());
                        }
                    }

					// TODO: chartColorStyle needs to be supported... line and fill styles should be created by following call (text, txtBody has to be added)
					// DMLHelper.createJsonFromShapeProperties(attrs, ser.getSpPr(), theme, chartStyle!=null?chartStyle.getChartArea():null, false);
					if (series.length() > 0) {
						final JSONObject insertChartDataSeriesOperation = new JSONObject(4);
						insertChartDataSeriesOperation.put("name", "insertChartDataSeries");
						insertChartDataSeriesOperation.put("start", position);
						insertChartDataSeriesOperation.put("series", seriesIndex);
						insertChartDataSeriesOperation.put("attrs", attrs);
						operationsArray.put(insertChartDataSeriesOperation);
					}

				}

				if (curved || valueLabel || varyColors != null || chartSer instanceof CTPieChart) {
					final JSONObject attrs = insertDrawingOperation.getJSONObject("attrs");
					final JSONObject chartProperties = attrs.getJSONObject("chart");

					if (chartSer instanceof CTPieChart) {
						chartProperties.put("rotation", ((CTPieChart) chartSer).getFirstSliceAng().getVal());
					}

					if (varyColors != null) {
						chartProperties.put("varyColors", varyColors.isVal());
					}

					if (valueLabel) {
						chartProperties.put("dataLabel", true);
					}

					if (curved) {
						chartProperties.put("curved", true);
					}
				}
			}
			addToOperations(chartPart, operationsArray, position);

			final CTTitle title = chart.getTitle();
			final JSONObject attrs = createJsonFromProperties(title);
			if (attrs != null) {
				final JSONObject op = new JSONObject(4);
				op.put("name", "setChartTitleAttributes");
				op.put("start", position);
				op.put("axis", "main");
				op.put("attrs", attrs);
				operationsArray.put(op);
			}

			final JSONObject legendAttrs = createLegendAttrs(chart);
			if (legendAttrs != null) {
				final JSONObject op = new JSONObject(3);
				op.put("name", "setChartLegendAttributes");
				op.put("start", position);
				op.put("attrs", legendAttrs);
				operationsArray.put(op);
			}
		}
	}

    private static JSONObject createLegendAttrs(CTChart chart) throws JSONException {
		final JSONObject ljset = new JSONObject();
		final JSONObject lj = new JSONObject();
		lj.put("pos", "off");

		final CTLegend legend = chart.getLegend();
		if (legend != null) {
		    //right is default, if legend exists
		    lj.put("pos", "right");
			if (legend.getLegendPos() != null && legend.getLegendPos().getVal() != null) {

				final String pos;
				switch (legend.getLegendPos().getVal()) {
				case R:
					pos = "right";
					break;
				case T:
					pos = "top";
					break;
				case B:
					pos = "bottom";
					break;
				case TR:
					pos = "topRight";
					break;
				default:
					pos = "left";
					break;
				}

				lj.put("pos", pos);

				final JSONObject character = DMLHelper.createJsonFromProperties(legend.getTxPr());
				if (character != null) {
					ljset.put("character", character);
				}
			}
		}
		ljset.put("legend", lj);

		return ljset;
	}

	private static Object getData(CTNumDataSource numDataSource) {
		if (numDataSource == null) {
			Logger.getAnonymousLogger().warning("CTNumDataSource is null");
			return null;
		}
		if (numDataSource.getNumRef() != null) {
			return numDataSource.getNumRef().getF();
		} else if (numDataSource.getNumLit() != null) {
			final CTNumData numLit = numDataSource.getNumLit();
			if (numLit != null) {
				final List<CTNumVal> numLitPtList = numLit.getPt();
				if (numLitPtList != null && !numLitPtList.isEmpty()) {
					final JSONArray values = new JSONArray();
					for (final CTNumVal numLitPt : numLitPtList) {
						if (numLitPt != null) {
							values.put(numLitPt.getV()); // TODO: getIdx
						} else {
							values.put((String) null);
						}
					}
					return values;
				}
			}
		}
		return null;
	}

	public static void applyChartSpaceProperties(JSONObject chartProperties, ChartWrapper chart, JSONObject fill) throws Exception {
		final CTPlotArea plotArea = chart.getPlotArea();
		final CTChartSpace chartSpace = chart.getChartSpace();

		if (chartProperties.has("type")) {
			// indirect call of getThemePart(createIfMissing=true) (Bug 46806)
			chart.getThemePart();

			List<IListSer> allCharts = plotArea.getAreaChartOrArea3DChartOrLineChart();

			//workaround for Bug 47969
			while (allCharts.size() > 1) {
				allCharts.remove(1);
			}

			final IListSer oldListSer = allCharts.get(0);
			final IListSer newListSer = createChartType(chartProperties);

			final Class<? extends ISerContent> type = getNewContentType(newListSer);

			if (newListSer != null) {
			    boolean noAxIdChart = newListSer instanceof NoAxIdChart;

				final List<CTUnsignedInt> axid = newListSer.getAxId();
				final List<CTUnsignedInt> oldAxid = oldListSer.getAxId();
				final List<IAxisDescription> axes = plotArea.getValAxOrCatAxOrDateAx();
				if (axid != null && oldAxid != null && !noAxIdChart) {
					axid.clear();
					axid.addAll(oldAxid);
				} else if (axid == null) {
					if (axes != null && oldAxid != null) {
						for (int i = axes.size()-1; i >= 0; i--)
                        {
                            IAxisDescription ax = axes.get(i);
                            CTUnsignedInt id = ax.getAxId();
                            for(CTUnsignedInt old : oldAxid){
                                if(id.getVal()==old.getVal()){
                                    axes.remove(i);
                                    break;
                                }
                            }

                        }
					}
				} else if (axes.size() == 0){
					axes.add(createCatAx(0,1));
					axes.add(createValAx(1,0));
				}

				if (noAxIdChart && axes != null) {
				    axes.clear();
				}

				allCharts.set(0, newListSer);

				newListSer.setVaryColors(oldListSer.getVaryColors());

				for (final ISerContent serContent : oldListSer.getSer()) {
					final ISerContent newContent;
					if (serContent.getClass().equals(type)) {
						newContent = serContent;
					} else {
						newContent = type.newInstance();
						newContent.setCat(serContent.getCat());
						newContent.setExtLst(serContent.getExtLst());
						newContent.setIdx(serContent.getIdx());
						newContent.setOrder(serContent.getOrder());
						newContent.setSpPr(serContent.getSpPr());
						newContent.setTx(serContent.getTx());
						newContent.setVal(serContent.getVal());
						newContent.setDLbls(serContent.getDLbls());

						if (newContent instanceof CTDoughnutChart) {
							handleStandardDonutChart((CTDoughnutChart) newContent, chartProperties);
						} else if (newContent instanceof CTPieChart) {
							handleStandardPieChart((CTPieChart) newContent, chartProperties);
						}

						setCurved(newContent, false);

					}
					newListSer.getSer().add(newContent);
				}

				if(axes.size()>0){
					if(newListSer instanceof CTBubbleChart || newListSer instanceof CTScatterChart){
						final IAxisDescription first = axes.get(0);
						if(first instanceof CTCatAx){
							final CTValAx valAx = createValAx(first.getAxId().getVal(), first.getCrossAx().getVal());
							copy(valAx, first);
							axes.set(0, valAx);
						}
					}else{
						final IAxisDescription first = axes.get(0);
						if(first instanceof CTValAx){
							final CTCatAx catAx = createCatAx(first.getAxId().getVal(), first.getCrossAx().getVal());
							copy(catAx, first);
							axes.set(0, catAx);
						}
					}
				}
			}
			refreshTitleRotations(chart);
		}

		final IListSer listSer = plotArea.getAreaChartOrArea3DChartOrLineChart().get(0);
		deleteEmptyContentLists(listSer);

		if (chartProperties.has("chartStyleId")) {
		    CTStyle chartStyle = chartSpace.getStyle();
			if (chartStyle == null) {
				chartStyle = new CTStyle();
				chartSpace.setStyle(chartStyle);
			}
			chartSpace.setSpPr(null);
			chartStyle.setVal((short) chartProperties.getInt("chartStyleId"));

			// FIXME: this should come with the series-operation
			// delete all shapeproperties, but keep the Line-NoFill-Property
			for (final ISerContent serContent : listSer.getSer()) {
				CTShapeProperties sppr = serContent.getSpPr();
				if (sppr != null ){
				    CTLineProperties ln = sppr.getLn();
				    if (null != ln && null != ln.getNoFill()) {
				        sppr = new CTShapeProperties();
                        sppr.setLn(ln);
                        serContent.setSpPr(sppr);
				    } else {
				        serContent.setSpPr(null);
				    }
				}
				final List<CTDPt> dpt = serContent.getDPt();
				if (dpt != null) {
					for (final CTDPt point : serContent.getDPt()) {
						point.setSpPr(null);
					}
				}
			}
		}
		if (chartProperties.has("stacking")) {
			final String stacking = chartProperties.optString("stacking", null);
			listSer.setStacking(stacking);
		}

		final String stacking = listSer.getStacking();
		if (listSer instanceof CTBarChart) {
			handleOverlap((CTBarChart) listSer, stacking);
		}


		if (chartProperties.has("curved")) {
			if (listSer instanceof CTLineChart || listSer instanceof CTScatterChart) {
				boolean curved = chartProperties.optBoolean("curved", false);
				setCurved(listSer, curved);
			}
		}
		if (chartProperties.has("varyColors")) {
			final CTBoolean vary = new CTBoolean();
			vary.setVal(chartProperties.optBoolean("varyColors", false));
			listSer.setVaryColors(vary);
		}

		if (chartProperties.has("dataLabel")) {
		    setDataLabel(listSer, chartProperties.optBoolean("dataLabel", false));
		}

		if (fill != null) {
			final JSONObject fillHolder = new JSONObject();
			fillHolder.put("fill", fill);

			plotArea.setSpPr(applyShapePropertiesFromJson(plotArea.getSpPr(), fillHolder));
			chartSpace.setSpPr(applyShapePropertiesFromJson(chartSpace.getSpPr(), fillHolder));

			chartSpace.setRoundedCorners(getBoolean(false));

		}

		if (listSer instanceof CTScatterChart) {
			handleStandardScatterChart((CTScatterChart) listSer, chartProperties);

		}

		// reset generic axis numbers
		// for Bug 46441
		if (chartProperties.has("stacking") || chartProperties.has("type")) {
			for (IAxisDescription axis : plotArea.getValAxOrCatAxOrDateAx()) {
				if (axis instanceof CTDateAx) {
					((CTDateAx)axis).setMajorUnit(null);
					((CTDateAx)axis).setMinorUnit(null);
					((CTDateAx)axis).setMajorTimeUnit(null);
					((CTDateAx)axis).setMinorTimeUnit(null);
				} else if (axis instanceof CTValAx) {
					((CTValAx)axis).setMajorUnit(null);
					((CTValAx)axis).setMinorUnit(null);
				}

				CTScaling scaling = axis.getScaling();
				if (null != scaling) {
					scaling.setMin(null);
					scaling.setMax(null);
				}
			}
		}
	}

    private static void copy(
		IAxisDescription target,
		IAxisDescription source)
	{
		target.setAxId(source.getAxId());
		target.setAxPos(source.getAxPos());
		target.setDelete(source.getDelete());
		target.setMajorGridlines(source.getMajorGridlines());
		target.setMajorTickMark(source.getMajorTickMark());
		target.setSpPr(source.getSpPr());
		target.setTickLblPos(source.getTickLblPos());
		target.setTitle(source.getTitle());
		target.setTxPr(source.getTxPr());
	}

    public static CTBoolean getBoolean(boolean value) {
		final CTBoolean res = new CTBoolean();
		res.setVal(value);
		return res;
	}

	private static void deleteEmptyContentLists(IListSer listSer) {
		for (final ISerContent serContent : listSer.getSer()) {
			if (hasEmptyContents(serContent.getCat())) {
				serContent.setCat(null);
			}
			if (hasEmptyContents(serContent.getTx())) {
				serContent.setTx(null);
			}
			if (serContent instanceof CTBubbleSer) {
				final CTBubbleSer serBub = ((CTBubbleSer) serContent);
				if (hasEmptyContents(serBub.getBubbleSize())) {
					serBub.setBubbleSize(null);
				}
			}
		}
	}

	private static boolean hasEmptyContents(CTNumDataSource data) {
		if (data == null) {
			return true;
		}
		if (data.getNumLit() != null || data.getNumRef() != null) {
			return false;
		}
		return true;
	}

	private static boolean hasEmptyContents(CTAxDataSource data) {
		if (data == null) {
			return true;
		}
		if (data.getMultiLvlStrRef() != null || data.getNumLit() != null || data.getNumRef() != null || data.getStrLit() != null || data.getStrRef() != null) {
			return false;
		}
		return true;
	}

	private static boolean hasEmptyContents(CTSerTx data) {
		if (data == null) {
			return true;
		}
		if (data.getStrRef() != null || data.getV() != null) {
			return false;
		}
		return true;
	}

	@SuppressWarnings("unchecked")
	private static Class<? extends ISerContent> getNewContentType(IListSer listSer) throws Exception {
		final XmlElements serType = listSer.getClass().getDeclaredField("ser").getAnnotation(XmlElements.class);
		return serType.value()[0].type();
	}

	private static CTUnsignedInt getUnsignedInt(int ptCount) {
		final CTUnsignedInt unsignedInt = Context.getDmlChartObjectFactory().createCTUnsignedInt();
		unsignedInt.setVal(ptCount);
		return unsignedInt;
	}

	private static String getRef(String sheetName, String ref) {
        if (!ref.contains("!")) {
            return sheetName + "!" + ref;
		}
		return ref;
	}

	// Idx and Order are required elements, so at least we have to create this
	// two elements for our new
	// inserted series... but we will correct indexes of other series, so that
	// they are properly increasing
	private static void updateIndexAndOrder(List<ISerContent> serContentList) {
		for (int i = 0; i < serContentList.size(); i++) {
			final ISerContent serCont = serContentList.get(i);
			serCont.setIdx(getUnsignedInt(i));
			serCont.setOrder(getUnsignedInt(i));
		}
	}

	public static void insertChartSpace(ChartWrapper chart, String sheetName, int series, JSONObject attrs)
		throws JSONException, InvalidFormatException, PartUnrecognisedException {

		final IListSer listSer = chart.getPlotArea().getAreaChartOrArea3DChartOrLineChart().get(0);
		final List<ISerContent> serContentList = listSer.getSer();
		final ISerContent serContent = listSer.createSer();

		disableInvertIfNegative(serContent);
        setCurved(serContent, isCurved(listSer));
        setDataLabel(serContent, isDataLabel(listSer));

		serContentList.add(series, serContent);
		setDataseriesAttrs(serContentList, serContent, sheetName, attrs);

	}

	private static void disableInvertIfNegative(ISerContent serContent) {
	    if (serContent instanceof CTBarSer) {
            ((CTBarSer)serContent).setInvertIfNegative(getBoolean(false));
        } else if (serContent instanceof CTBubbleSer) {
            ((CTBubbleSer)serContent).setInvertIfNegative(getBoolean(false));
        }
    }

    public static void setDataSeriesAttributes(ChartWrapper chart, String sheetName, int series, JSONObject attrs)
        throws JSONException, InvalidFormatException, PartUnrecognisedException {

        List<IListSer> allCharts = chart.getPlotArea().getAreaChartOrArea3DChartOrLineChart();
        final IListSer listSer = allCharts.get(0);

        final List<ISerContent> serContentList = listSer.getSer();
        final ISerContent serContent = serContentList.get(series);

        disableInvertIfNegative(serContent);
        setCurved(serContent, isCurved(listSer));
        setDataLabel(serContent, isDataLabel(listSer));

        setDataseriesAttrs(serContentList, serContent, sheetName, attrs);

    }

    private static void setDataseriesAttrs(List<ISerContent> serContentList, ISerContent serContent, String sheetName, JSONObject attrs)
        throws JSONException, InvalidFormatException, PartUnrecognisedException {

        if (attrs != null) {
            final JSONObject attrSeries = attrs.optJSONObject("series");
            if (attrSeries != null) {

                // title
                if (attrSeries.has("title")) {
                    CTSerTx serTx = null;
                    final Object title = attrSeries.get("title");
                    if (title != null) {
                        serTx = Context.getDmlChartObjectFactory().createCTSerTx();

                        if (title instanceof String)
                        {
                            final CTStrRef strRef = serTx.getStrRef(true);
                            strRef.setF(getRef(sheetName, (String) title));

                            // empty cache for Net-Charts! (Bug 49821)
                            strRef.setStrCache(Context.getDmlChartObjectFactory().createCTStrData());
                        }
                        else if (title instanceof JSONArray && !((JSONArray) title).isEmpty())
                        {
                            final Object o = ((JSONArray) title).get(0);
                            if (o instanceof String)
                            {
                                serTx.setV((String) o);
                            }
                        }
                        else
                        {
                            Logger.getAnonymousLogger().warning("Cant handle Chart Title " + title);
                            serTx = null;
                        }
                    }
                    serContent.setTx(serTx);
                }
                // names
                if (attrSeries.has("names")) {
                    CTAxDataSource cat = null;
                    final Object names = attrSeries.get("names");
                    if (names != null) {
                        cat = Context.getDmlChartObjectFactory().createCTAxDataSource();

                        if (names instanceof String)
                        {
                            final CTStrRef strRef = cat.getStrRef(true);
                            strRef.setF(getRef(sheetName, (String) names));

                            // empty cache for Net-Charts! (Bug 49821)
                            strRef.setStrCache(Context.getDmlChartObjectFactory().createCTStrData());
                        }
                        else if (names instanceof JSONArray)
                        {
                            final JSONArray titleArray = (JSONArray) names;
                            final CTStrData strData = cat.getStrLit(true);
                            strData.setPtCount(getUnsignedInt(titleArray.length()));
                            for (int i = 0; i < titleArray.length(); i++)
                            {
                                final CTStrVal strVal = Context.getDmlChartObjectFactory().createCTStrVal();
                                strVal.setV(titleArray.optString(i, ""));
                                strData.getPt().add(strVal);
                            }
                        }
                        else
                        {
                            Logger.getAnonymousLogger().warning("Cant handle Chart Names " + names);
                            cat = null;
                        }
                    }
                    serContent.setCat(cat);
                }
                // values
                if (attrSeries.has("values")) {
                    CTNumDataSource val = null;
                    final Object values = attrSeries.get("values");
                    if (values != null) {
                        val = Context.getDmlChartObjectFactory().createCTNumDataSource();
                        insertSeriesAttribute(sheetName, val, values);
                    }
                    serContent.setVal(val);
                }

                // bubbles
                if (attrSeries.has("bubbles") && serContent instanceof CTBubbleSer) {
                    CTNumDataSource bub = null;
                    final Object bubbles = attrSeries.opt("bubbles");
                    if (bubbles != null) {
                        bub = Context.getDmlChartObjectFactory().createCTNumDataSource();
                        insertSeriesAttribute(sheetName, bub, bubbles);
                    }
                    ((CTBubbleSer) serContent).setBubbleSize(bub);
                }
                if (serContent instanceof CTScatterSer) {
                    handleStandardScatterSer((CTScatterSer) serContent);
                }
            }

            serContent.setSpPr(applyShapePropertiesFromJson(serContent.getSpPr(), attrs));

            if (attrs.has("fill")) {
                if (serContent instanceof CTLineSer || serContent instanceof CTScatterSer) {
                    CTMarker marker = Context.getDmlChartObjectFactory().createCTMarker();

                    marker.setSpPr(new CTShapeProperties());
                    marker.getSpPr().setNoFill(new CTNoFillProperties());
                    marker.getSpPr().setLn(new CTLineProperties());
                    marker.getSpPr().getLn().setNoFill(new CTNoFillProperties());

                    DMLHelper.applyFillPropertiesFromJson(marker.getSpPr(), attrs, null, null);

                    if (serContent instanceof CTLineSer) {
                        ((CTLineSer) serContent).setMarker(marker);
                    } else if (serContent instanceof CTScatterSer) {
                        ((CTScatterSer) serContent).setMarker(marker);
                    }
                }
            }

            updateIndexAndOrder(serContentList);
        }
    }

    private static CTShapeProperties applyShapePropertiesFromJson(CTShapeProperties shape, JSONObject attrs)
        throws InvalidFormatException, PartUnrecognisedException, JSONException {

        // workaround for Bug 43983 & 44012
        // normal fill behavior works fine, but not for Dark Chart, then they will be invisible,
        // but without any fill info everything looks fine

        if (attrs.has("fill")) {
            JSONObject fill = attrs.optJSONObject("fill");
            if (fill != null) {
                if (fill.optString("type", "").equals("solid")) {
                    JSONObject color = fill.optJSONObject("color");
                    if (color == null || color.optString("type", "").equals("auto")) {
                        //attrs.remove("fill");
                        fill.remove("color");
                    }
                }
            }
            // fix for Bug 48209
            handlePlaceholderColor(attrs, "fill");

        }
        if (attrs.has("line")) {
            handlePlaceholderColor(attrs, "line");
        }

        return DMLHelper.applyShapePropertiesFromJson(shape, attrs, null, null, true);
    }

    private static void handlePlaceholderColor(JSONObject attrs, String key) throws JSONException {
        JSONObject shape = attrs.optJSONObject(key);
        if (null != shape) {
            if (shape.optString("type", "").equals("none")) {
                shape.remove("color");
            } else {
                JSONObject color = shape.optJSONObject("color");
                if (null != color) {
                    String type = color.optString("type");
                    if (null != type && type.equals("scheme")) {
                        String value = color.optString("value");
                        if (null != value && value.equals("phClr")) {
                            //attrs.remove(key);
                            shape.remove("color");
                        }
                    }
                }
            }
        }
    }

	private static void insertSeriesAttribute(String sheetName, CTNumDataSource source, Object attribute) {
		if (attribute instanceof String) {
			final CTNumRef numRef = source.getNumRef(true);
			numRef.setF(getRef(sheetName, (String) attribute));

            // empty cache for Net-Charts! (Bug 49821)
            numRef.setNumCache(Context.getDmlChartObjectFactory().createCTNumData());
		} else if (attribute instanceof JSONArray) {
			final JSONArray bubblesArray = (JSONArray) attribute;
			final CTNumData numData = Context.getDmlChartObjectFactory().createCTNumData();
			source.setNumLit(numData);
			numData.setPtCount(getUnsignedInt(bubblesArray.length()));
			for (int i = 0; i < bubblesArray.length(); i++) {
				final CTNumVal numVal = Context.getDmlChartObjectFactory().createCTNumVal();
				numVal.setIdx(i);
				numVal.setV(bubblesArray.optString(i, ""));
				numData.getPt().add(numVal);
			}
		}
	}

	public static void deleteChartSpace(ChartWrapper chart, int series) {

		final IListSer listSer = chart.getPlotArea().getAreaChartOrArea3DChartOrLineChart().get(0);
		final List<ISerContent> serContentList = listSer.getSer();
		serContentList.remove(series);
		updateIndexAndOrder(serContentList);
	}

	public static CTChartSpace createChartSpace(JSONObject chartAttrs) {
		final org.docx4j.dml.chart.ObjectFactory chartObjectFactory = Context.getDmlChartObjectFactory();
		final CTChartSpace chartSpace = chartObjectFactory.createCTChartSpace();
		final CTChart chart = chartObjectFactory.createCTChart();
		chartSpace.setChart(chart);
		final CTPlotArea plotArea = chartObjectFactory.createCTPlotArea();
		chart.setPlotArea(plotArea);
		final CTLegend legend = chartObjectFactory.createCTLegend();
		chart.setLegend(legend);
		final CTLegendPos legendPos = chartObjectFactory.createCTLegendPos();
		legend.setLegendPos(legendPos);
		legendPos.setVal(STLegendPos.R);
		final CTBoolean booleanValue = chartObjectFactory.createCTBoolean();
		booleanValue.setVal(true);
		chart.setPlotVisOnly(booleanValue);
		final IListSer listSer = createChartType(chartAttrs);
		plotArea.getAreaChartOrArea3DChartOrLineChart().add(listSer);

		/*
		 * CTCatAx CTSerAx CTValAx CTDateAx
		 */
		final List<IAxisDescription> axList = plotArea.getValAxOrCatAxOrDateAx();

		if (listSer instanceof CTBubbleChart || listSer instanceof CTScatterChart)
		{
			axList.add(createValAx(0,1));
			axList.add(createValAx(1,0));
		}
		else
		{
			axList.add(createCatAx(0,1));
			axList.add(createValAx(1,0));
		}
		return chartSpace;
	}

	private static CTCatAx createCatAx(long axe, long crossId) {
		final org.docx4j.dml.chart.ObjectFactory chartObjectFactory = Context.getDmlChartObjectFactory();
		final CTCatAx catAx = chartObjectFactory.createCTCatAx();
		final CTUnsignedInt axId = chartObjectFactory.createCTUnsignedInt();
		axId.setVal(axe);
		catAx.setAxId(axId);
		final CTScaling scaling = chartObjectFactory.createCTScaling();
		final CTOrientation orientation = chartObjectFactory.createCTOrientation();
		orientation.setVal(STOrientation.MIN_MAX);
		scaling.setOrientation(orientation);
		catAx.setScaling(scaling);
		final CTAxPos axPos = chartObjectFactory.createCTAxPos();
		axPos.setVal(STAxPos.B);
		catAx.setAxPos(axPos);
		final CTUnsignedInt crossAx = chartObjectFactory.createCTUnsignedInt();
		crossAx.setVal(crossId);
		catAx.setCrossAx(crossAx);
        catAx.setNumFmt(null);
		catAx.setSpPr(getHiddenShapeProperties());
		return catAx;
	}

	private static CTValAx createValAx(long axe, long crossId) {
		final org.docx4j.dml.chart.ObjectFactory chartObjectFactory = Context.getDmlChartObjectFactory();
		final CTValAx valAx = chartObjectFactory.createCTValAx();
		final CTUnsignedInt axId = chartObjectFactory.createCTUnsignedInt();
		axId.setVal(axe);
		valAx.setAxId(axId);
		final CTScaling scaling = chartObjectFactory.createCTScaling();
		final CTOrientation orientation = chartObjectFactory.createCTOrientation();
		orientation.setVal(STOrientation.MIN_MAX);
		scaling.setOrientation(orientation);
		valAx.setScaling(scaling);
		final CTAxPos axPos = chartObjectFactory.createCTAxPos();
		axPos.setVal(STAxPos.L);
		valAx.setAxPos(axPos);
		final CTUnsignedInt crossAx = chartObjectFactory.createCTUnsignedInt();
		crossAx.setVal(crossId);
		valAx.setCrossAx(crossAx);
        valAx.setNumFmt(null);
		valAx.setSpPr(getHiddenShapeProperties());
		return valAx;
	}

	private static void createAxId(List<CTUnsignedInt> axId) {
		final org.docx4j.dml.chart.ObjectFactory chartObjectFactory = Context.getDmlChartObjectFactory();
		final CTUnsignedInt id0 = chartObjectFactory.createCTUnsignedInt();
		final CTUnsignedInt id1 = chartObjectFactory.createCTUnsignedInt();
		id0.setVal(0);
		id1.setVal(1);
		axId.add(id0);
		axId.add(id1);
	}

	// TODO: creator for pie3d needs to be implemented...

	private static IListSer createChartType(JSONObject chartAttrs) {

		IListSer listSer = null;
		final ObjectFactory chartObjectFactory = Context.getDmlChartObjectFactory();
		final String chartType = chartAttrs != null ? chartAttrs.optString("type", "bar2d") : "bar2d";

		if (chartType.equals("bar3d")) {
			listSer = createBarChartType(chartType, chartAttrs, chartObjectFactory);
		}
		else if (chartType.equals("column2d")) {
			listSer = createBarChartType(chartType, chartAttrs, chartObjectFactory);
		}
		else if (chartType.equals("column3d")) {
			listSer = createBarChartType(chartType, chartAttrs, chartObjectFactory);
		}
		else if (chartType.equals("line2d")) {
			listSer = chartObjectFactory.createCTLineChart();
		}
		else if (chartType.equals("line3d")) {
			listSer = chartObjectFactory.createCTLine3DChart();
		}
		else if (chartType.equals("area2d")) {
			listSer = chartObjectFactory.createCTAreaChart();
		}
		else if (chartType.equals("area3d")) {
			listSer = chartObjectFactory.createCTArea3DChart();
		}
		else if (chartType.equals("radar2d")) {
			CTRadarChart radar = chartObjectFactory.createCTRadarChart();
			radar.setRadarStyle(new CTRadarStyle());
			radar.getRadarStyle().setVal(STRadarStyle.STANDARD);
			listSer = radar;
		}
		else if (chartType.equals("scatter2d")) {
			final CTScatterChart scatter = chartObjectFactory.createCTScatterChart();
			handleStandardScatterChart(scatter, chartAttrs);
			listSer = scatter;
		}
		else if (chartType.equals("bubble2d")) {
			listSer = chartObjectFactory.createCTBubbleChart();
		}
		else if (chartType.equals("pie2d")) {
			final CTPieChart pie = chartObjectFactory.createCTPieChart();
			handleStandardPieChart(pie, chartAttrs);
			listSer = pie;
		}
		else if (chartType.equals("donut2d")) {
			final CTDoughnutChart donut = chartObjectFactory.createCTDoughnutChart();
			handleStandardDonutChart(donut, chartAttrs);
			listSer = donut;
		}
		else { // bar2d -> default
			listSer = createBarChartType(chartType, chartAttrs, chartObjectFactory);
		}
		final List<CTUnsignedInt> axid = listSer.getAxId();
		if (axid != null) {
			createAxId(listSer.getAxId());
		}
		listSer.setStacking("standard");

		return listSer;
	}


	private static void handleStandardScatterSer(CTScatterSer serContent) {
		CTShapeProperties shape = serContent.getSpPr();
		if (null == shape) {
		    shape = new CTShapeProperties();
	        serContent.setSpPr(shape);
		}
		CTLineProperties line = shape.getLn();
		if (null == line) {
		    line = new CTLineProperties();
		    shape.setLn(line);
		    line.setW(19050);
		    line.setCap(STLineCap.RND);
	        line.setRound(new CTLineJoinRound());
	        shape.getEffectLst();
		}

	}

	private static void handleStandardScatterChart(CTScatterChart scatter, JSONObject chartAttrs) {
		if(scatter.getScatterStyle() == null || scatter.getScatterStyle().getVal() == null){
			final CTScatterStyle style = new CTScatterStyle();
			style.setVal(STScatterStyle.LINE_MARKER);
			scatter.setScatterStyle(style);
		}

		for (final ISerContent cont : scatter.getSer()) {
			handleStandardScatterSer((CTScatterSer) cont);
		}
	}

	private static void handleStandardPieChart(CTPieChart pie, JSONObject chartAttrs) {
		pie.setFirstSliceAng(new CTFirstSliceAng());
	}

	private static void handleStandardDonutChart(CTDoughnutChart donut, JSONObject chartAttrs) {
		final CTHoleSize holeSize = new CTHoleSize();
		holeSize.setVal((short) 60);
		donut.setHoleSize(holeSize);
	}

	private static IListSer createBarChartType(String chartType, JSONObject chartAttrs, ObjectFactory chartObjectFactory) {
		IListSer listSer = null;
		String barDir = null;
		if (chartType.equals("bar3d")) {
			listSer = chartObjectFactory.createCTBar3DChart();
			barDir = "bar";
		}
		else if (chartType.equals("column2d")) {
			listSer = chartObjectFactory.createCTBarChart();
			barDir = "col";
		}
		else if (chartType.equals("column3d")) {
			listSer = chartObjectFactory.createCTBar3DChart();
			barDir = "col";
		}
		else {
			listSer = chartObjectFactory.createCTBarChart();
			barDir = "bar";
		}
		listSer.setBarDirection(barDir);


		if (listSer instanceof CTBarChart) {
			String stacking = null;
			if (chartAttrs != null) {
				stacking = chartAttrs.optString("stacking", null);
			}
			handleOverlap((CTBarChart) listSer, stacking);
		}

		return listSer;
	}

	private static void handleOverlap(CTBarChart barSer, String stacking) {
		final CTGapAmount gap = new CTGapAmount();
		gap.setVal(150);
		final CTOverlap overlap;
		if (stacking == null || stacking.equals("standard") || stacking.equals("clustered")) {
			overlap = null;
		} else {
			overlap = new CTOverlap();
			overlap.setVal((byte) 100);
		}
		barSer.setGapWidth(gap);
		barSer.setOverlap(overlap);

	}

	public static void getChartType(ChartWrapper chart, HashMap<String, Object> map) {
		final CTPlotArea plotArea = chart.getPlotArea();
		if (plotArea == null) {
			return;
		}
		final List<IListSer> chartList = plotArea.getAreaChartOrArea3DChartOrLineChart();
		if (chartList.isEmpty()) {
			return;
		}

		String chartType = null;
		final IListSer listSer = chartList.get(0);
		if (listSer instanceof CTBubbleChart) {
			chartType = "bubble2d";
		}
		else if (listSer instanceof CTRadarChart) {
			chartType = "radar2d";
		}
		else if (listSer instanceof CTSurface3DChart) {
			chartType = "bar2d";
		}
		else if (listSer instanceof CTArea3DChart) {
			chartType = "area3d";
		}
		else if (listSer instanceof CTOfPieChart) {
			chartType = "bar2d";
		}
		else if (listSer instanceof CTAreaChart) {
			chartType = "area2d";
		}
		else if (listSer instanceof CTBarChart) {
			chartType = handleBarDir(listSer);
		}
		else if (listSer instanceof CTSurfaceChart) {
			chartType = "bar2d";
		}
		else if (listSer instanceof CTLine3DChart) {
			chartType = "line3d";
		}
		else if (listSer instanceof CTDoughnutChart) {
			chartType = "donut2d";
		}
		else if (listSer instanceof CTLineChart) {
			chartType = "line2d";
		}
		else if (listSer instanceof CTScatterChart) {
			chartType = "scatter2d";
		}
		else if (listSer instanceof CTBar3DChart) {
			chartType = handleBarDir(listSer);
		}
		else if (listSer instanceof CTPieChart) {
			chartType = "pie2d";
		}
		else if (listSer instanceof CTPie3DChart) {
			chartType = "pie3d";
		}
		else if (listSer instanceof CTStockChart) {
			chartType = "bar2d";
			map.put("chartGroup", "clustered");
		}
		if (chartType != null) {
			map.put("chartType", chartType);
		}
		final String chartGroup = listSer.getStacking();
		if (chartGroup != null) {
			map.put("chartGroup", chartGroup);
		}
	}

	private static String handleBarDir(IListSer listSer) {
		final CTBarDir dir;
		final String view;
		if (listSer instanceof CTBar3DChart) {
			view = "3d";
			dir = ((CTBar3DChart) listSer).getBarDir();
		} else {
			view = "2d";
			dir = ((CTBarChart) listSer).getBarDir();
		}
		final String dirString;
		switch (dir.getVal()) {
		case BAR:
			dirString = "bar";
			break;
		default:
			dirString = "column";
			break;
		}
		return dirString + view;
	}

	// workaround for Bug 43959
	public static CTTextBody setTextBodyFromAttrs(CTTextBody textBody, JSONObject attrs)
			throws JSONException {
		JSONObject character = attrs.optJSONObject("character");
		if (character == null || character.length() == 0) {
			//nothing
		} else {
			if (textBody == null) {
				textBody = new CTTextBody();
				DMLHelper.handleStandardTypes(textBody);
			}
			setCharacterFromAttrs(textBody, attrs);
		}
		return textBody;
	}

	public static void setLegendFromAttrs(ChartWrapper chart, JSONObject attrs) throws JSONException {
	    final CTChart ctChart = chart.getChartSpace().getChart();
		final String pos = attrs.getJSONObject("legend").optString("pos", "off");
		if (pos.equals("off")) {
			ctChart.setLegend(null);
		} else {
			CTLegend legend = ctChart.getLegend();
			if (legend == null) {
				legend = new CTLegend();
				ctChart.setLegend(legend);
			}

			legend.setTxPr(setTextBodyFromAttrs(legend.getTxPr(), attrs));

			legend.setOverlay(new CTBoolean());
			legend.getOverlay().setVal(false);

			CTLegendPos legendPos = legend.getLegendPos();
			if (legendPos == null) {
				legendPos = new CTLegendPos();
				legend.setLegendPos(legendPos);
			}

			final STLegendPos stpos;
			if (pos.equals("left")) {
				stpos = STLegendPos.L;
			} else if (pos.equals("top")) {
				stpos = STLegendPos.T;
			} else if (pos.equals("right")) {
				stpos = STLegendPos.R;
			} else if (pos.equals("bottom")) {
				stpos = STLegendPos.B;
			} else if (pos.equals("topRight")) {
				stpos = STLegendPos.TR;
			} else {
				Logger.getAnonymousLogger().warning("No STLegendPost-enum found for " + pos);
				stpos = STLegendPos.L;
			}
			legendPos.setVal(stpos);
		}
	}

	private static boolean isCurved(IListSer chartSer) {
		if (chartSer instanceof CTScatterChart){
			CTScatterStyle style = ((CTScatterChart) chartSer).getScatterStyle();
			if (style != null && (style.getVal() == STScatterStyle.SMOOTH || style.getVal() == STScatterStyle.SMOOTH_MARKER)){
				return true;
			}
		}else if(chartSer instanceof CTLineChart){
			CTBoolean smooth = ((CTLineChart) chartSer).getSmooth();
			if(smooth != null && smooth.isVal()){
				return true;
			}
		}
		for (ISerContent serContent : chartSer.getSer()) {
			if(isCurved(serContent)){
				return true;
			}
		}
		return false;
	}

	private static boolean isCurved(ISerContent serContent) {
		if (serContent instanceof CTLineSer) {
			final CTBoolean smooth = ((CTLineSer) serContent).getSmooth();
			//no smooth in line chart means it is activated!!!
			if (smooth == null || smooth.isVal()) {
				return true;
			}
		} else if (serContent instanceof CTScatterSer) {
			final CTBoolean smooth = ((CTScatterSer) serContent).getSmooth();
			//no smooth in line chart means it is activated!!!
			if (smooth == null || smooth.isVal()) {
				return true;
			}
		}
		return false;
	}


	private static void setCurved(ISerContent serContent, boolean curved) {
		final CTBoolean smooth = new CTBoolean();
		smooth.setVal(curved);

		if (serContent instanceof CTLineSer) {
			((CTLineSer) serContent).setSmooth(smooth);
		} else if (serContent instanceof CTScatterSer) {
			((CTScatterSer) serContent).setSmooth(smooth);
		}

	}

	private static void setCurved(IListSer listSer, boolean curved) {
		final CTBoolean smooth = new CTBoolean();
		smooth.setVal(curved);
		if (listSer instanceof CTLineChart){
			((CTLineChart)listSer).setSmooth(smooth);
		}else if (listSer instanceof CTScatterChart){
			final CTScatterStyle style = new CTScatterStyle();
			if(smooth.isVal()){
				style.setVal(STScatterStyle.SMOOTH_MARKER);
			}else{
				style.setVal(STScatterStyle.LINE_MARKER);
			}
			((CTScatterChart)listSer).setScatterStyle(style);
		}
		for (final ISerContent serContent : listSer.getSer()) {
			setCurved(serContent, curved);
		}
	}

    private static boolean isDataLabel(IListSer chartSer){
        CTDLbls lbl = chartSer.getDLbls();
        if (lbl != null && lbl.getShowVal() != null && lbl.getShowVal().isVal())
        {
            return true;
        }

        for (ISerContent serContent : chartSer.getSer())
        {
            if (isDataLabel(serContent))
            {
                return true;
            }
        }
        return false;
    }

    private static boolean isDataLabel(ISerContent serContent){
        CTDLbls lbl = serContent.getDLbls();
        if (lbl != null && lbl.getShowVal() != null && lbl.getShowVal().isVal())
        {
            return true;
        }
        return false;
    }

    private static void setDataLabel(IListSer chartSer, boolean labeled) {
        CTDLbls lbl = chartSer.getDLbls();
        if (lbl == null){
            lbl = newDataLabel();
            chartSer.setDLbls(lbl);
        }
        lbl.setShowVal(getBoolean(labeled));

        for (ISerContent serContent : chartSer.getSer())
        {
            setDataLabel(serContent, labeled);
        }

    }

    private static void setDataLabel(ISerContent serContent, boolean labeled) {
        CTDLbls lbl = serContent.getDLbls();
        if (lbl == null){
            lbl = newDataLabel();
            serContent.setDLbls(lbl);
        }
        lbl.setShowVal(getBoolean(labeled));
    }

    private static CTDLbls newDataLabel(){
        final CTDLbls lbl = new CTDLbls();
        lbl.setShowLegendKey(getBoolean(false));
        lbl.setShowCatName(getBoolean(false));
        lbl.setShowSerName(getBoolean(false));
        lbl.setShowPercent(getBoolean(false));
        lbl.setShowBubbleSize(getBoolean(false));
        lbl.setShowLeaderLines(getBoolean(false));

        CTShapeProperties spr = lbl.getSpPr();
        if (spr == null) {
            spr = new CTShapeProperties();
            lbl.setSpPr(spr);
            spr.setNoFill(new CTNoFillProperties());
            final CTLineProperties ln = new CTLineProperties();
            ln.setNoFill(new CTNoFillProperties());
            spr.setLn(ln);

            spr.setEffectLst(new CTEffectList());
        }
        CTTextBody txPr = lbl.getTxPr();
        if(txPr==null){
            txPr = new CTTextBody();
            lbl.setTxPr(txPr);
            DMLHelper.handleStandardTypes(txPr);
        }

        final CTTextParagraph p = new CTTextParagraph();
        final CTTextParagraphProperties prop = p.getPPr(true);
        final CTTextCharacterProperties def = prop.getDefRPr(true);

        def.setSz(1000);

        final List<Object> paras = txPr.getContent();
        paras.clear();

        paras.add(p);

        return lbl;
    }

    public static JSONObject createJsonFromChartColor(CTColorStyle color)
        throws JSONException {

        JSONObject jsonChartColor = null;
        if(color!=null) {
            jsonChartColor = new JSONObject();
            jsonChartColor.put("id", color.getId());
            jsonChartColor.put("meth", color.getMeth());
            final List<Object> schemeClrList = color.getEGColorChoice();;
            if(schemeClrList!=null) {
                final JSONArray jsonArray = new JSONArray();
                for(int i=0; i<schemeClrList.size(); i++) {
                	final ColorChoiceListEntry listEntry = new ColorChoiceListEntry(schemeClrList, i);
                	if(listEntry.getSchemeClr()!=null) {
                		final JSONObject jsonColor = DMLHelper.createJsonColorFromSolidColorFillProperties(listEntry);
                		if(jsonColor!=null) {
                			jsonArray.put(jsonColor);
                		}
                	}
                }
                jsonChartColor.put("schemeClr", jsonArray);
            }
            final List<CTColorStyleVariation> variationList = color.getVariation();
            if(variationList!=null ) {

            	/* TODO: variationList needs to be implemented...

                final JSONArray jsonArray = new JSONArray();
                for(final CTColorStyleVariation schemeClr:variationList) {
                    final JSONObject resColor = new JSONObject();
                	createJsonFromSchemeClr(resColor, schemeClr);
    				if (resColor != null){
    					jsonArray.put(resColor);
    				}
                }
    			if (!jsonArray.isEmpty()){
    				jsonChartColor.put("variation", jsonArray);
    			}
    			*/
            }
        }
        return jsonChartColor!=null&&!jsonChartColor.isEmpty()?jsonChartColor:null;
    }

    public static void refreshTitleRotations(
        ChartWrapper chart)
    {
        CTPlotArea plotArea = chart.getPlotArea();
    	final List<IAxisDescription> axes = plotArea.getValAxOrCatAxOrDateAx();
    	final IListSer ser = plotArea.getAreaChartOrArea3DChartOrLineChart().get(0);
    	final String dir = ser.getBarDirection();
    	for (int i = 0; i < axes.size(); i++)
    	{
    		boolean rotate = false;
    		if (i == 1)
    		{
    			rotate = true;
    		}
    		if (dir == null || dir.equals("col"))
    		{
    			// normal
    		}
    		else
    		{
    			rotate = !rotate;
    		}
    		int rot = 0;
    		if (rotate)
    		{
    			rot = -5400000;
    		}
    		final IAxisDescription axis = axes.get(i);
    		final CTTitle title = axis.getTitle();

            if (title != null)
    		{
                if (title.getTx() != null && title.getTx().getRich() != null && title.getTx().getRich().getBodyPr() != null) {
                    final CTTx tx = title.getTx();
                    final CTTextBody rich = tx.getRich();
                    final CTTextBodyProperties body = rich.getBodyPr();
                    body.setRot(rot);

                }

    			final CTTextBody txPr = title.getTxPr();
                if (null != txPr && null != txPr.getBodyPr()) {
    			    final CTTextBodyProperties body2 = txPr.getBodyPr();
    			    body2.setRot(rot);
    			}
    		}

    		final CTTextBody lbl = axis.getTxPr();
    		if (lbl != null && lbl.getBodyPr() != null)
    		{
    			final CTTextBodyProperties body = lbl.getBodyPr();
    			body.setRot(0);
    		}
    	}
    }

    static final void getAxeTitleOperation(IAxisDescription axe, JSONArray operations, String axisParam, JSONArray position) throws JSONException {
    	final CTTitle title = axe.getTitle();

    	final JSONObject attrs = createJsonFromProperties(title);
    	if (attrs != null) {

    		final JSONObject op = new JSONObject();
    		op.put("name", "setChartTitleAttributes");
    		op.put("start", position);
    		op.put("axis", axisParam);
    		op.put("attrs", attrs);
    		operations.put(op);
    	}

    }

    public static void setChartTitleAttributes(ChartWrapper chart, String axis, JSONObject attrs) throws JSONException {
    	final IAxisDescription ax = getAxis(chart, axis);
    	if (ax == null) {
    		return;
    	}

    	CTTitle title = ax.getTitle();
    	if (title == null) {
    		title = new CTTitle();
    		ax.setTitle(title);
    	}
    	setTitleFromAttrs(title, attrs);
    	refreshTitleRotations(chart);
    }

    static final void getGridOperation(ChartWrapper chart, IAxisDescription axe, JSONArray operations, String axisParam, JSONArray position) throws Exception {

    	final CTChartLines grid = axe.getMajorGridlines();
    	if (grid != null) {
    		final JSONObject gridOp = new JSONObject();
    		final CTShapeProperties shape = grid.getSpPr();
    		if (shape != null) {
    			DMLHelper.createJsonFromShapeProperties(gridOp, shape, chart.getThemePart(), chart.getChartStyle().getGridlineMajor(), chart.getChartPart(), false, true);
    		}
    		if (!gridOp.has("line")) {
    		    //FIXME: nessecary?
    			gridOp.put("line",  makeStandardShapeType());
    		}
    		final JSONObject op = new JSONObject();
    		op.put("name", "setChartGridlineAttributes");
    		op.put("start", position);
    		op.put("axis", axisParam);
    		op.put("attrs", gridOp);
    		operations.put(op);
    	}
    }

    static IAxisDescription getAxis(ChartWrapper chart, String axis) {
        CTPlotArea plotArea = chart.getPlotArea();

        try {
        	// workaround for Bug 46915
			if (plotArea.getAreaChartOrArea3DChartOrLineChart().get(0) instanceof NoAxIdChart) {
				return null;
			}
		} catch (Exception e) {
			e.printStackTrace();
		}

    	final int index;
    	if (axis.equals("x")) {
    		index = 0;
    	} else if (axis.equals("y")) {
    		index = 1;
    	} else if (axis.equals("z")) {
    		index = 2;
    	} else {
    		return null;
    	}
    	final List<IAxisDescription> axes = plotArea.getValAxOrCatAxOrDateAx();

    	IAxisDescription res = null;
    	if (index < axes.size()) {
    		res = axes.get(index);
    	}

    	if (res == null) {
    		if (axis.equals("x")) {
    			final CTCatAx cat = createCatAx(0,1);
    			cat.setMinorGridlines(getHiddenChartLines());
    			cat.setMinorTickMark(getHiddenTickMark());
                cat.setNumFmt(null);
    			res = cat;
    		} else {
    			final CTValAx val = createValAx(1,0);
    			val.setMinorGridlines(getHiddenChartLines());
    			val.setMinorTickMark(getHiddenTickMark());
                val.setNumFmt(null);
    			res = val;
    		}
    		res.setAxId(getUnsignedInt(index));
    		axes.add(index, res);
    	}
        // fix for Bug 49940
        if (res.getDelete() != null && res.getDelete().isVal()) {
    		hideAxis(res);
        } else {
            // workaround for Bug 49959
            res.setDelete(getBoolean(false));
    	}

    	return res;
    }

    static void hideAxis(
    	IAxisDescription axis)
    {
    	final CTBoolean delete = getBoolean(false);
    	axis.setDelete(delete);

    	axis.setSpPr(getHiddenShapeProperties());

    	axis.setMajorGridlines(getHiddenChartLines());
    	axis.setTickLblPos(getHiddenLblPos());

    	axis.setTitle(null);
    	axis.setTxPr(null);

    }

    public static JSONObject createJsonFromProperties(CTTitle title) throws JSONException {
    	if (title == null) {
    		return null;
    	}
    	final CTTx tx = title.getTx();
    	if (tx == null) {
    		return null;
    	}

    	final CTTextBody rich = tx.getRich();
    	final CTStrRef ref = tx.getStrRef();

    	if (null == ref && null == rich) {
    		return null;
    	}

    	CTTextCharacterProperties characterProb = null;

    	final JSONObject attrs = new JSONObject();
    	final JSONObject textObj = new JSONObject();
    	attrs.put("text", textObj);

    	if (rich != null) {

	    	final StringBuffer buf = new StringBuffer();

			final List<Object> paras = rich.getContent();
			for (int i = 0; i < paras.size(); i++) {
				final CTTextParagraph p = (CTTextParagraph)paras.get(i);
				final List<Object> textes = p.getContent();
				for (final Object text : textes) {
					if(text instanceof CTRegularTextRun) {
						addTo((CTRegularTextRun)text, buf);
					}
					else if (text instanceof CTTextField) {
						addTo((CTTextField)text, buf);
					}
					else if (text instanceof CTTextLineBreak) {
						addTo((CTTextLineBreak)text, buf);
					}

					if (characterProb == null) {
						characterProb = ((ITextCharacterProperties)text).getRPr(false);
					}

				}
				if (i < paras.size() - 1) {
					buf.append("\n");
				}
			}

	    	final JSONArray link = new JSONArray();
	    	link.put(buf.toString());
	    	textObj.put("link", link);
	    }

    	if (ref != null) {
    		textObj.put("link", ref.getF());
    	}

    	if (characterProb != null) {
    		DMLHelper.createJsonFromTextCharacterProperties(attrs, characterProb, null);
    	}
    	return attrs;
    }


    private static JSONArray attrHasTitle(JSONObject attrs) throws JSONException {
        final JSONObject text = attrs.optJSONObject("text");
        if(text==null){
            return null;
        }

        final Object olink = text.opt("link");

        if (olink == null || !(olink instanceof JSONArray))
        {

            if (olink == null || olink.equals(JSONObject.NULL))
            {
                return null;
            }
            else
            {
                Logger.getAnonymousLogger().warning("title text has no correct content " + text.opt("link"));
                return null;
            }
        }

        final JSONArray links = (JSONArray) olink;
        if (links.length() > 0)
        {
            return links;
        }
        return null;
    }

    private static String attrHasRef(JSONObject attrs) throws JSONException {
        final JSONObject text = attrs.optJSONObject("text");
        if(text==null){
            return null;
        }

        final Object olink = text.opt("link");

        if (olink instanceof String)
        {
            return (String) olink;
        }

        return null;
    }

    public static void setTitleFromAttrs(CTTitle title, JSONObject attrs) throws JSONException {
        handleStandardTypes(title);

        JSONArray links = attrHasTitle(attrs);
        String ref = attrHasRef(attrs);

        if (links == null && ref == null) {

            title.setTx(null);
            title.setTxPr(setTextBodyFromAttrs(title.getTxPr(), attrs));
        } else {

            final CTTx tx = title.getTx();

            final CTTextBody rich = tx.getRich();
            final CTTextBodyProperties body = rich.getBodyPr();
            body.setRot(0);

            final List<Object> paras = rich.getContent();
            paras.clear();

            //delete cell ref, fix for Bug 47761
            tx.setStrRef(null);

            if (ref != null && ref.length() > 0) {

                CTStrRef strRef = new CTStrRef();
                strRef.setF(ref);
                tx.setStrRef(strRef);
                tx.setRich(null);
                return;
            }

            title.setTxPr(null);

            if (links != null && links.length() > 0) {
                for (int i = 0; i < links.length(); i++)

                {
                    final CTTextParagraph p = Context.getDmlObjectFactory().createCTTextParagraph();
                    setCharacterFromAttrs(p, attrs);
                    paras.add(p);
                    final List<Object> textHolder = p.getContent();
                    textHolder.clear();

                    final String link = links.getString(i);
                    final String[] textes = link.split("\n");
                    for (int j = 0; j < textes.length; j++)
                    {
                        final CTRegularTextRun reg = Context.getDmlObjectFactory().createCTRegularTextRun();
                        reg.setRPr(setCharacterFromAttrs(reg.getRPr(false), attrs));
                        reg.setT(textes[j]);
                        textHolder.add(reg);
                        if (j < textes.length - 1)
                        {
                            final CTTextLineBreak lb = Context.getDmlObjectFactory().createCTTextLineBreak();
                            lb.setRPr(setCharacterFromAttrs(lb.getRPr(false), attrs));
                            textHolder.add(lb);
                        }
                    }
                }
            }
        }
    }

    static void handleStandardTypes(CTTitle title) {
    	CTTx tx = title.getTx();
    	if (tx == null) {
    		tx = new CTTx();
    		title.setTx(tx);
    	}
    	CTTextBody rich = tx.getRich();
    	if (rich == null) {
    		rich = new CTTextBody();
    		tx.setRich(rich);
    	}
    	DMLHelper.handleStandardTypes(rich);

    	CTLayout layout = title.getLayout();
    	if (layout == null) {
    		layout = new CTLayout();
    		title.setLayout(layout);
    	}
    	CTBoolean overlay = title.getOverlay();
    	if (overlay == null) {
    		overlay = new CTBoolean();
    		title.setOverlay(overlay);
    	}
    	overlay.setVal(false);
    	CTShapeProperties spr = title.getSpPr();
    	if (spr == null) {
    		spr = getHiddenShapeProperties();
    	}

    	final List<Object> ps = rich.getContent();
    	ps.clear();

    	final CTTextParagraph p = new CTTextParagraph();
    	p.getPPr(true);
    	ps.add(p);

    }

    public static void addToOperations(ChartWrapper chart, JSONArray operations, JSONArray position) throws Exception {
        CTPlotArea plotArea = chart.getPlotArea();
    	final List<IAxisDescription> axes = plotArea.getValAxOrCatAxOrDateAx();
    	if (axes != null && axes.size() > 0) {
    		for (int i = 0; i < axes.size(); i++) {
    			final IAxisDescription axe = axes.get(i);

    			final String axisParam;
    			if (i == 0) {
    				axisParam = "x";
    			} else if (i == 1) {
    				axisParam = "y";
    			} else if (i == 2) {
    				axisParam = "z";
    			} else {
    				return;
    			}

    			getAxeOperation(chart, axe, operations, axisParam, position);
    			getGridOperation(chart, axe, operations, axisParam, position);
    		}
    	}
    }

    static final void getAxeOperation(ChartWrapper chart, IAxisDescription axe, JSONArray operations, String axisParam, JSONArray position) throws Exception {

    	final JSONObject axis = new JSONObject();

    	/* disabled for Bug 46441, if we want to support scaling,w e also must support major & minor units!
        final CTScaling scaling = axe.getScaling();
        if(scaling!=null) {
            if (scaling.getMin()!=null) {
                axis.put("min", scaling.getMin().getVal());
            }
            if(scaling.getMax()!=null) {
                axis.put("max", scaling.getMax().getVal());
            }
        }
        */
        JSONObject character = null;
    	final JSONObject axisOp = new JSONObject();
    	final CTBoolean del = axe.getDelete();
    	if (del != null && del.isVal()) {
    		final JSONObject line = new JSONObject();
    		line.put("type", "none");
    		axisOp.put("line", line);
    	} else {
    		final CTShapeProperties shape = axe.getSpPr();
    		if (shape != null) {
                DMLHelper.createJsonFromShapeProperties(axisOp, shape, chart.getThemePart(), chart.getChartStyle().getSeriesAxis(), chart.getChartPart(), false, true);
    		}
    		if (!axisOp.has("line")) {
    		    //FIXME: nessecary?
    			axisOp.put("line",  makeStandardShapeType());
    		}

    		final CTTickLblPos label = axe.getTickLblPos();
    		if (label != null && label.getVal() != STTickLblPos.NONE) {
    			axis.put("label", true);
    			character = DMLHelper.createJsonFromProperties(axe.getTxPr());
    		}

    		getAxeTitleOperation(axe, operations, axisParam, position);
    	}

    	if (axis.length() > 0) {
    		axisOp.put("axis", axis);
    	}
    	if (character != null) {
    		axisOp.put("character", character);
    	}


    	if (axisOp.length() > 0) {
    		final JSONObject op = new JSONObject();
    		op.put("name", "setChartAxisAttributes");
    		op.put("start", position);
    		op.put("axis", axisParam);
    		op.put("attrs", axisOp);
    		operations.put(op);
    	}
    }

    public static void setAxisAttributes(ChartWrapper chart, String axis, JSONObject attrs)
    	throws JSONException, InvalidFormatException, PartUnrecognisedException {

    	final IAxisDescription ax = getAxis(chart, axis);
    	if (ax == null) {
    		return;
    	}
    	final CTShapeProperties shape = DMLHelper.applyShapePropertiesFromJson(ax.getSpPr(), attrs, null, null, true);
    	ax.setSpPr(shape);
    	CTTickMark tickmark = ax.getMajorTickMark();
    	if (tickmark == null)
    	{
    		tickmark = new CTTickMark();
    		ax.setMajorTickMark(tickmark);
    	}
    	if (shape.getLn() != null && shape.getLn().getNoFill() != null)
    	{
    		tickmark.setVal(STTickMark.NONE);
    	}
    	else
    	{
    		tickmark.setVal(STTickMark.OUT);
    	}

    	final JSONObject a = attrs.optJSONObject("axis");
    	if (a != null) {
    		CTTickLblPos lblPos = ax.getTickLblPos();
    		if (lblPos == null) {
    			lblPos = new CTTickLblPos();
    			ax.setTickLblPos(lblPos);
    		}
    		final boolean label = a.optBoolean("label", false);

    		if (label) {
    			lblPos.setVal(STTickLblPos.LOW);

    			ax.setTxPr(setTextBodyFromAttrs(ax.getTxPr(), attrs));
    		} else {
    			lblPos.setVal(STTickLblPos.NONE);
    		}
    	}
    }

    public static void setChartGridlineAttributes(ChartWrapper chart, String axis, JSONObject attrs)
    	throws JSONException, InvalidFormatException, PartUnrecognisedException {

    	final IAxisDescription ax = getAxis(chart, axis);
    	if (ax == null) {
    		return;
    	}

    	CTChartLines grids = ax.getMajorGridlines();
    	if (grids == null) {
    		grids = new CTChartLines();
    		ax.setMajorGridlines(grids);
    	}
    	grids.setSpPr(DMLHelper.applyShapePropertiesFromJson(grids.getSpPr(), attrs, null, null, true));
    }

    static CTChartLines getHiddenChartLines()
    {
    	final CTChartLines line = new CTChartLines();
    	line.setSpPr(getHiddenShapeProperties());
    	return line;
    }

    static CTTickLblPos getHiddenLblPos()
    {
    	final CTTickLblPos lbl = new CTTickLblPos();
    	lbl.setVal(STTickLblPos.NONE);
    	return lbl;
    }

    static CTTickMark getHiddenTickMark()
    {
    	final CTTickMark mark = new CTTickMark();
    	mark.setVal(STTickMark.NONE);
    	return mark;
    }


	public static void setCharacterFromAttrs(CTTextBody txPr, JSONObject attrs) throws JSONException {
		final List<Object> ps = txPr.getContent();
		ps.clear();
		final CTTextParagraph p = Context.getDmlObjectFactory().createCTTextParagraph();
		ps.add(p);
		setCharacterFromAttrs(p, attrs);
	}

	public static void setCharacterFromAttrs(CTTextParagraph p, JSONObject attrs) throws JSONException {

		final JSONObject character = attrs.optJSONObject("character");

		if (character != null) {
			final CTTextParagraphProperties prop = p.getPPr(true);
			prop.setDefRPr(setCharacterFromAttrs(prop.getDefRPr(false), attrs));
		}
	}

	public static CTTextCharacterProperties setCharacterFromAttrs(CTTextCharacterProperties t, JSONObject attrs)
			throws JSONException {
		final JSONObject character = attrs.optJSONObject("character");

		if (character != null) {
			if (t == null) {
				t = Context.getDmlObjectFactory().createCTTextCharacterProperties();
			}
			DMLHelper.applyTextCharacterPropertiesFromJson(t, attrs, null);

		}
		return t;
	}

	public static CTShapeProperties getHiddenShapeProperties()
	{
		final CTShapeProperties shape = new CTShapeProperties();
		shape.setNoFill(new CTNoFillProperties());
		final CTLineProperties line = new CTLineProperties();
		line.setNoFill(new CTNoFillProperties());
		shape.setLn(line);
		return shape;
	}

    //FIXME: nessecary?
    static JSONObject makeStandardShapeType() throws JSONException {
    	final JSONObject shape = new JSONObject();
    	shape.put("type", "solid");
    	final JSONObject color = new JSONObject();
    	color.put("type", "auto");
    	shape.put("color", color);
    	return shape;
    }


    public static void addTo(CTRegularTextRun textRun, StringBuffer stringBuffer) {
    	stringBuffer.append(textRun.getT());
    }

    public static void addTo(CTTextField textRun, StringBuffer stringBuffer) {
    	stringBuffer.append(textRun.getT());
    }

    public static void addTo(CTTextLineBreak textRun, StringBuffer stringBuffer) {
    	stringBuffer.append("\n");
    }

}
