/*
 *
 *    OPEN-XCHANGE legal information
 *
 *    All intellectual property rights in the Software are protected by
 *    international copyright laws.
 *
 *
 *    In some countries OX, OX Open-Xchange, open xchange and OXtender
 *    as well as the corresponding Logos OX Open-Xchange and OX are registered
 *    trademarks.
 *    The use of the Logos is not covered by the GNU General Public License.
 *    Instead, you are allowed to use these Logos according to the terms and
 *    conditions of the Creative Commons License, Version 2.5, Attribution,
 *    Non-commercial, ShareAlike, and the interpretation of the term
 *    Non-commercial applicable to the aforementioned license is published
 *    on the web site http://www.open-xchange.com/EN/legal/index.html.
 *
 *    Please make sure that third-party modules and libraries are used
 *    according to their respective licenses.
 *
 *    Any modifications to this package must retain all copyright notices
 *    of the original copyright holder(s) for the original code used.
 *
 *    After any such modifications, the original and derivative code shall remain
 *    under the copyright of the copyright holder(s) and/or original author(s)per
 *    the Attribution and Assignment Agreement that can be located at
 *    http://www.open-xchange.com/EN/developer/. The contributing author shall be
 *    given Attribution for the derivative code and a license granting use.
 *
 *     Copyright (C) 2016 OX Software GmbH
 *     Mail: info@open-xchange.com
 *
 *
 *     This program is free software; you can redistribute it and/or modify it
 *     under the terms of the GNU General Public License, Version 2 as published
 *     by the Free Software Foundation.
 *
 *     This program is distributed in the hope that it will be useful, but
 *     WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *     or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 *     for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc., 59
 *     Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

package com.openexchange.office.realtime.osgi;

import java.util.concurrent.atomic.AtomicBoolean;

import org.apache.commons.logging.Log;

import com.openexchange.capabilities.CapabilityService;
import com.openexchange.config.ConfigurationService;
import com.openexchange.context.ContextService;
import com.openexchange.conversion.simple.SimplePayloadConverter;
import com.openexchange.documentconverter.IDocumentConverter;
import com.openexchange.file.storage.composition.IDBasedFileAccessFactory;
import com.openexchange.file.storage.composition.IDBasedFolderAccessFactory;
import com.openexchange.filemanagement.DistributedFileManagement;
import com.openexchange.filemanagement.ManagedFileManagement;
import com.openexchange.folderstorage.FolderService;
import com.openexchange.jslob.JSlobService;
import com.openexchange.office.IResourceProvider;
import com.openexchange.office.backup.manager.DocumentBackupController;
import com.openexchange.office.hazelcast.doc.DocSaveStateDirectory;
import com.openexchange.office.hazelcast.doc.DocumentDirectory;
import com.openexchange.office.realtime.impl.ConnectionBackgroundProcessorManager;
import com.openexchange.office.realtime.impl.ConnectionComponent;
import com.openexchange.office.realtime.impl.FailSafeSaveProcessor;
import com.openexchange.office.realtime.impl.calc.CalcComponent;
import com.openexchange.office.realtime.impl.presentation.PresentationComponent;
import com.openexchange.office.realtime.impl.text.TextComponent;
import com.openexchange.office.realtime.tools.JSON2MessageData;
import com.openexchange.office.realtime.tools.MessageData2JSON;
import com.openexchange.osgi.HousekeepingActivator;
import com.openexchange.realtime.Component;
import com.openexchange.realtime.cleanup.GlobalRealtimeCleanup;
import com.openexchange.realtime.cleanup.RealtimeJanitor;
import com.openexchange.threadpool.ThreadPoolService;
import com.openexchange.timer.TimerService;
import com.openexchange.user.UserService;

//The activator does the service wiring, as usual.
public class ConnectionActivator extends HousekeepingActivator {

    @Override
    protected Class<?>[] getNeededServices() {
        return new Class<?>[] {
            com.openexchange.office.odf.Importer.class,
            com.openexchange.office.odf.Exporter.class,
            com.openexchange.office.ooxml.pptx.Importer.class,
            com.openexchange.office.ooxml.pptx.Exporter.class,
            com.openexchange.office.ooxml.docx.Importer.class,
            com.openexchange.office.ooxml.docx.Exporter.class,
            com.openexchange.office.ooxml.xlsx.Importer.class,
            com.openexchange.office.ooxml.xlsx.Exporter.class,
            IResourceProvider.class,
            ConfigurationService.class,
            CapabilityService.class,
            IDBasedFileAccessFactory.class,
            IDBasedFolderAccessFactory.class,
            FolderService.class,
            ManagedFileManagement.class,
            DistributedFileManagement.class,
            UserService.class,
            TimerService.class,
            com.openexchange.sessiond.SessiondService.class,
            GlobalRealtimeCleanup.class,
            JSlobService.class,
            ThreadPoolService.class,
            ContextService.class,
            DocumentDirectory.class,
            DocSaveStateDirectory.class,
            DocumentBackupController.class
        };
    }

    /* (non-Javadoc)
     * @see com.openexchange.osgi.HousekeepingActivator#getOptionalServices()
     */
    @Override
    protected Class<?>[] getOptionalServices() {
        return new Class<?>[] {
            IDocumentConverter.class
        };
    }

    /* (non-Javadoc)
     * @see com.openexchange.osgi.DeferredActivator#startBundle()
     */
    @Override
    protected void startBundle() throws Exception {
        Services.setServiceLookup(this);

        LOG.debug("starting bundle: " + getClass().getCanonicalName());

        // Register the components
        registerService(Component.class, textComponent = new TextComponent(this));
        registerService(Component.class, calcComponent = new CalcComponent(this));
        registerService(Component.class, presComponent = new PresentationComponent(this));
        // Register the components as RealtimeJanitor that can be tracked via OSGI and used when a cleanup is needed
        registerService(RealtimeJanitor.class, textComponent, textComponent.getServiceProperties());
        registerService(RealtimeJanitor.class, calcComponent, calcComponent.getServiceProperties());
        registerService(RealtimeJanitor.class, presComponent, presComponent.getServiceProperties());
        // You have to provide these converters if you want to put non JSON classes into
        // Payload trees.
        // Also, if you want to receive non-json data, which is usually very convenient, you
        // can declare that a payload with a certain namespace and element name should always be
        // Converted into a specific format. For that see the PayloadTreeConverter interface
        registerService(SimplePayloadConverter.class, new MessageData2JSON());
        registerService(SimplePayloadConverter.class, new JSON2MessageData());

        // Register document specific background processor dependent on
        // configuration entries.
        final ConfigurationService configuration = this.getService(com.openexchange.config.ConfigurationService.class);
        boolean backgroundSave = configuration.getBoolProperty("io.ox/office//module/backgroundSaveDoc", true);
        LOG.debug("Background document save = " + backgroundSave);
        if (backgroundSave) {
            m_backgroundProcesses = new ConnectionBackgroundProcessorManager(this);
            final FailSafeSaveProcessor failSafeSaveProcessor = new FailSafeSaveProcessor(this);
            m_backgroundProcesses.registerProcessor(failSafeSaveProcessor);
            // only add connection pool if we know that at least one processor has been added
            m_backgroundProcesses.addConnectionPool(textComponent);
            m_backgroundProcesses.addConnectionPool(calcComponent);
            m_backgroundProcesses.addConnectionPool(presComponent);
            m_backgroundProcesses.start();
        }

        openTrackers();
        isStopped.set(false);
    }

    @Override
    protected void stopBundle() throws Exception {
        if (isStopped.compareAndSet(false, true)) {

            LOG.debug("stopping bundle: " + getClass().getCanonicalName());

            if (null != m_backgroundProcesses) {
                m_backgroundProcesses.shutdown();
            }
            if (null != textComponent) {
                textComponent.shutdown();
            }
            if (null != calcComponent) {
                calcComponent.shutdown();
            }
            if (null != presComponent) {
                presComponent.shutdown();
            }

            Services.setServiceLookup(null);

            cleanUp();
        }
    }

    @Override
    /**
     * Handles the re-availability of a needed service. The specific activator may
     * decide which actions are further done dependent on given service's class.
     */
    protected void handleAvailability(Class<?> clazz) {
        if (allAvailable()) {
            LOG.debug("ConnectionActivator regained needed service " + clazz.getSimpleName() + ". Going to start bundle.");
            try {
                startBundle();
            } catch (Exception e) {
                LOG.error("Error while starting bundle.", e);
            }
        }
    }

    @Override
    /**
     * Called by the HousekeepingActivator when a needed service will be
     * unavailable after handleUnavailability() returns.
     *
     * @see com.openexchange.osgi.HousekeepingActivator#handleUnavailability()
     */
    protected void handleUnavailability(final Class<?> clazz) {
        LOG.warn("ConnectionActivator is handling unavailibility of needed service " + clazz.getSimpleName() + ". Going to stop bundle.");
        try {
            this.stopBundle();
        } catch (Exception e) {
            LOG.error("Error while stopping bundle.", e);
        }
    }

    // - Members ---------------------------------------------------------------

    @SuppressWarnings("deprecation")
    private final Log LOG = com.openexchange.log.Log.loggerFor(ConnectionActivator.class);

    private ConnectionBackgroundProcessorManager m_backgroundProcesses = null;

    private final AtomicBoolean isStopped = new AtomicBoolean(true);

    private ConnectionComponent textComponent = null;

    private ConnectionComponent calcComponent = null;

    private ConnectionComponent presComponent = null;
}
